import logging
import json
import datetime
import uuid

from flask import g, request, url_for
from flask_restful import Api
from flask_socketio import SocketIO
from http import HTTPStatus
from sqlalchemy import desc
from sqlalchemy.sql import visitors

from vantage6.common.globals import STRING_ENCODING, NodePolicy
from vantage6.common.task_status import TaskStatus, has_task_finished
from vantage6.common.encryption import DummyCryptor
from vantage6.backend.common import get_server_url
from vantage6.server import db
from vantage6.server.algo_store_communication import request_algo_store
from vantage6.server.permission import (
    RuleCollection,
    Scope as S,
    PermissionManager,
    Operation as P,
)
from vantage6.server.resource import only_for, ServicesResources, with_user
from vantage6.server.resource.common.output_schema import (
    TaskSchema,
    TaskWithResultSchema,
    TaskWithRunSchema,
    TaskWithRunAndResultSchema,
)
from vantage6.server.resource.common.input_schema import TaskInputSchema
from vantage6.backend.common.resource.pagination import Pagination
from vantage6.server.resource.event import kill_task


module_name = __name__.split(".")[-1]
log = logging.getLogger(module_name)


def setup(api: Api, api_base: str, services: dict) -> None:
    """
    Setup the task resource.

    Parameters
    ----------
    api : Api
        Flask restful api instance
    api_base : str
        Base url of the api
    services : dict
        Dictionary with services required for the resource endpoints
    """
    path = "/".join([api_base, module_name])
    log.info(f'Setting up "{path}" and subdirectories')

    api.add_resource(
        Tasks,
        path,
        endpoint="task_without_id",
        methods=("GET", "POST"),
        resource_class_kwargs=services,
    )
    api.add_resource(
        Task,
        path + "/<int:id>",
        endpoint="task_with_id",
        methods=("GET", "DELETE"),
        resource_class_kwargs=services,
    )
    api.add_resource(
        TaskStatusEndpoint,
        path + "/<int:task_id>/status",
        endpoint="task_status",
        methods=("GET",),
        resource_class_kwargs=services,
    )


# -----------------------------------------------------------------------------
# Permissions
# -----------------------------------------------------------------------------
def permissions(permissions: PermissionManager) -> None:
    """
    Define the permissions for this resource.

    Parameters
    ----------
    permissions : PermissionManager
        Permission manager instance to which permissions are added
    """
    add = permissions.appender(module_name)

    add(scope=S.GLOBAL, operation=P.VIEW, description="view any task")
    add(
        scope=S.COLLABORATION,
        operation=P.VIEW,
        assign_to_container=True,
        assign_to_node=True,
        description="view tasks of your collaborations",
    )
    add(
        scope=S.ORGANIZATION,
        operation=P.VIEW,
        description="view tasks that your organization initiated",
    )
    add(scope=S.OWN, operation=P.VIEW, description="view tasks that you initiated")

    add(scope=S.GLOBAL, operation=P.CREATE, description="create a new task")
    add(
        scope=S.COLLABORATION,
        operation=P.CREATE,
        description=(
            "create a new task for collaborations in which your organization "
            "participates with"
        ),
    )

    add(scope=S.GLOBAL, operation=P.DELETE, description="delete a task")
    add(
        scope=S.COLLABORATION,
        operation=P.DELETE,
        description="delete a task from your collaborations",
    )
    add(
        scope=S.ORGANIZATION,
        operation=P.DELETE,
        description=(
            "delete a task from a collaboration in which your organization "
            "participates with"
        ),
    )
    add(scope=S.OWN, operation=P.DELETE, description="delete tasks that you created")


# ------------------------------------------------------------------------------
# Resources / API's
# ------------------------------------------------------------------------------
task_schema = TaskSchema()
task_run_schema = TaskWithRunSchema()
task_result_schema = TaskWithResultSchema()
task_result_run_schema = TaskWithRunAndResultSchema()

task_input_schema = TaskInputSchema()


class TaskBase(ServicesResources):
    def __init__(self, socketio, storage_adapter, mail, api, permissions, config):
        super().__init__(socketio, storage_adapter, mail, api, permissions, config)
        self.r: RuleCollection = getattr(self.permissions, module_name)
        # permissions for the run resource are also relevant for the task
        # resource as they are sometimes included
        self.r_run: RuleCollection = getattr(self.permissions, "run")

    def _select_schema(self) -> TaskSchema:
        """
        Select the schema to use for serialization.

        Returns
        -------
        TaskSchema
            Schema to use for serialization
        """
        if self.is_included("runs") and self.is_included("results"):
            return task_result_run_schema
        elif self.is_included("runs"):
            return task_run_schema
        elif self.is_included("results"):
            return task_result_schema
        else:
            return task_schema


class Tasks(TaskBase):
    @only_for(("user", "node", "container"))
    def get(self):
        """List tasks
        ---
        description: >-
          Returns a list of tasks.\n

          ### Permission Table\n
          |Rule name|Scope|Operation|Assigned to node|Assigned to container|
          Description|\n
          |--|--|--|--|--|--|\n
          |Task|Global|View|❌|❌|View any task|\n
          |Task|Collaboration|View|✅|✅|View any task in your collaborations|
          \n
          |Task|Organization|View|❌|❌|View any task that your organization
          created|\n
          |Task|Own|View|❌|❌|View any task that you created|\n

          Accessible to users.

        parameters:
          - in: query
            name: init_org_id
            schema:
              type: int
            description: The organization id of the origin of the request
          - in: query
            name: init_user_id
            schema:
              type: int
            description: The user id of the user that started the task
          - in: query
            name: collaboration_id
            schema:
              type: int
            description: The collaboration id to which the task belongs
          - in: query
            name: study_id
            schema:
              type: int
            description: The study id to which the task belongs
          - in: query
            name: is_user_created
            schema:
              type: int
            description: >-
              If larger than 0, returns tasks created by a user (top-level
              tasks). If equal to 0, returns subtask created by an algorithm.
              If not specified, both are returned.
          - in: query
            name: image
            schema:
              type: str
            description: >-
              (Docker) image name which is used in the task. Name to match
              with a LIKE operator. \n
              * The percent sign (%) represents zero, one, or multiple
              characters\n
              * underscore sign (_) represents one, single character
          - in: query
            name: parent_id
            schema:
              type: int
            description: The id of the parent task
          - in: query
            name: job_id
            schema:
              type: int
            description: The run id that belongs to the task
          - in: query
            name: store_id
            schema:
              type: int
            description: The algorithm store ID from which the algorithm was retrieved
          - in: query
            name: name
            schema:
              type: str
            description: >-
              Name to match with a LIKE operator. \n
              * The percent sign (%) represents zero, one, or multiple
              characters\n
              * underscore sign (_) represents one, single character
          - in: query
            name: description
            schema:
              type: string
            description: >-
              Description to match with a LIKE operator. \n
              * The percent sign (%) represents zero, one, or multiple
              characters\n
              * underscore sign (_) represents one, single character
          - in: query
            name: database
            schema:
              type: string
            description: >-
              Database description to match with a LIKE operator. \n
              * The percent sign (%) represents zero, one, or multiple
              characters\n
              * underscore sign (_) represents one, single character
          - in: query
            name: run_id
            schema:
              type: int
            description: A run id that belongs to the task
          - in: query
            name: include
            schema:
              type: array
              items:
                type: string
            description: Include 'results' to include the task's results,
              'runs' to include details on algorithm runs. For including
               multiple, do either `include=x,y` or `include=x&include=y`.
          - in: query
            name: status
            schema:
              type: string
            description: Filter by task status, e.g. 'active' for active
              tasks, 'completed' for finished or 'crashed' for failed tasks.
          - in: query
            name: page
            schema:
              type: integer
            description: Page number for pagination (default=1)
          - in: query
            name: per_page
            schema:
              type: integer
            description: Number of items per page (default=10)
          - in: query
            name: sort
            schema:
              type: string
            description: >-
              Sort by one or more fields, separated by a comma. Use a minus
              sign (-) in front of the field to sort in descending order.

        responses:
          200:
            description: Ok
          400:
            description: Non-allowed or wrong parameter values
          401:
            description: Unauthorized

        security:
            - bearerAuth: []

        tags: ["Task"]
        """
        q = g.session.query(db.Task)
        args = request.args

        auth_org_id = self.obtain_organization_id()

        # check permissions and apply filter if neccassary
        if not self.r.v_glo.can():
            if self.r.v_col.can():
                q = (
                    q.join(db.Collaboration)
                    .join(db.Organization)
                    .filter(db.Collaboration.organizations.any(id=auth_org_id))
                )
            elif self.r.v_org.can():
                q = q.join(db.Organization).filter(db.Task.init_org_id == auth_org_id)
            elif self.r.v_own.can():
                q = q.filter(db.Task.init_user_id == g.user.id)
            else:
                return {
                    "msg": "You lack the permission to do that!"
                }, HTTPStatus.UNAUTHORIZED
        # if results are included, check permissions on results
        if self.is_included("results"):
            max_scope_task = self.r.get_max_scope(P.VIEW)
            if not self.r_run.has_at_least_scope(max_scope_task, P.VIEW):
                max_scope_run = self.r_run.get_max_scope(P.VIEW)
                return {
                    "msg": "You cannot view the results of all tasks, as you "
                    f"are allowed to view tasks with scope {max_scope_task} "
                    f"but you can only view results with scope {max_scope_run}"
                }, HTTPStatus.UNAUTHORIZED

        if "collaboration_id" in args:
            collaboration_id = int(args["collaboration_id"])
            if not self.r.can_for_col(P.VIEW, collaboration_id):
                return {
                    "msg": "You lack the permission to view tasks "
                    f"from collaboration {collaboration_id}!"
                }, HTTPStatus.UNAUTHORIZED
            # dont join collaboration table if it is already joined
            # FIXME refactor this after moving to SQLAlchemy 2.0
            has_already_joined_collab = False
            for visitor in visitors.iterate(q.statement):
                if (
                    visitor.__visit_name__ == "table"
                    and visitor.name == "collaboration"
                ):
                    has_already_joined_collab = True
            if not has_already_joined_collab:
                q = q.join(db.Collaboration)
            q = q.filter(db.Collaboration.id == collaboration_id)

        if "init_org_id" in args:
            init_org_id = int(args["init_org_id"])
            if not self.r.allowed_for_org(P.VIEW, init_org_id):
                return {
                    "msg": "You lack the permission to view tasks "
                    f"from organization id={init_org_id}!"
                }, HTTPStatus.UNAUTHORIZED
            q = q.filter(db.Task.init_org_id == init_org_id)

        if "init_user_id" in args:
            init_user_id = int(args["init_user_id"])
            init_user = db.User.get(init_user_id)
            if not init_user:
                return {
                    "msg": f"User id={init_user_id} does not " "exist!"
                }, HTTPStatus.BAD_REQUEST
            elif not self.r.allowed_for_org(P.VIEW, init_user.organization_id) and not (
                self.r.v_own.can() and g.user and init_user.id == g.user.id
            ):
                return {
                    "msg": "You lack the permission to view tasks "
                    f"from user id={init_user_id}!"
                }, HTTPStatus.UNAUTHORIZED
            q = q.filter(db.Task.init_user_id == init_user_id)

        if "study_id" in args:
            study_id = int(args["study_id"])
            study = db.Study.get(study_id)
            if not study:
                return {
                    "msg": f"Study id={study_id} does not exist!"
                }, HTTPStatus.BAD_REQUEST
            elif not self.r.can_for_col(P.VIEW, study.collaboration_id):
                return {
                    "msg": "You lack the permission to view tasks "
                    f"from collaboration id={study.collaboration_id} that the study "
                    f"with id={study.id} belongs to!"
                }, HTTPStatus.UNAUTHORIZED
            q = q.filter(db.Task.study_id == study_id)

        if "parent_id" in args:
            parent_id = int(args["parent_id"])
            parent = db.Task.get(parent_id)
            if not parent:
                return {
                    "msg": f'Parent task id={args["parent_id"]} does not ' "exist!"
                }, HTTPStatus.BAD_REQUEST
            elif not self.r.can_for_col(P.VIEW, parent.collaboration_id):
                return {
                    "msg": "You lack the permission to view tasks "
                    "from the collaboration that the task with parent_id="
                    f"{parent.collaboration_id} belongs to!"
                }, HTTPStatus.UNAUTHORIZED
            q = q.filter(db.Task.parent_id == int(parent_id))

        if "job_id" in args:
            job_id = int(args["job_id"])
            task_in_job = (
                q.session.query(db.Task).filter(db.Task.job_id == job_id).first()
            )
            if not task_in_job:
                return {
                    "msg": f'Job id={args["job_id"]} does not exist!'
                }, HTTPStatus.BAD_REQUEST
            elif not self.r.can_for_col(P.VIEW, task_in_job.collaboration_id):
                return {
                    "msg": "You lack the permission to view tasks "
                    "from the collaboration that the task with job_id="
                    f"{task_in_job.collaboration_id} belongs to!"
                }, HTTPStatus.UNAUTHORIZED
            q = q.filter(db.Task.job_id == job_id)

        for param in ["name", "image", "description", "status"]:
            if param in args:
                q = q.filter(getattr(db.Task, param).like(args[param]))

        if "run_id" in args:
            run_id = int(args["run_id"])
            run = db.Run.get(run_id)
            if not run:
                return {
                    "msg": f'Run id={args["run_id"]} does not exist!'
                }, HTTPStatus.BAD_REQUEST
            elif not self.r.can_for_col(P.VIEW, run.collaboration_id):
                return {
                    "msg": "You lack the permission to view tasks "
                    "from the collaboration that the run with id="
                    f"{run.collaboration_id} belongs to!"
                }, HTTPStatus.UNAUTHORIZED
            q = q.join(db.Run).filter(db.Run.id == run_id)

        if "store_id" in args:
            store_id = int(args["store_id"])
            store = db.AlgorithmStore.get(store_id)
            if not store:
                return {
                    "msg": f"Algorithm store id={store_id} does not exist!"
                }, HTTPStatus.BAD_REQUEST
            q = q.filter(db.Task.algorithmstore_id == store_id)

        if "database" in args:
            q = q.join(db.TaskDatabase).filter(
                db.TaskDatabase.database == args["database"]
            )

        if "is_user_created" in args:
            try:
                user_created = int(args["is_user_created"])
                if user_created == 0:
                    q = q.filter(db.Task.parent_id.isnot(None))
                else:
                    q = q.filter(db.Task.parent_id.is_(None))
            except ValueError:
                return {
                    "msg": (
                        "Invalid value for 'is_user_created' provided: "
                        f"'{args['is_user_created']}'. Should be an integer."
                    )
                }, HTTPStatus.BAD_REQUEST

        # order to get latest task first
        q = q.order_by(desc(db.Task.id))

        # paginate tasks
        try:
            page = Pagination.from_query(q, request, db.Task)
        except (ValueError, AttributeError) as e:
            return {"msg": str(e)}, HTTPStatus.INTERNAL_SERVER_ERROR

        # serialization schema
        schema = self._select_schema()

        return self.response(page, schema)

    @only_for(("user", "container"))
    def post(self):
        """Adds new computation task
        ---
        description: >-
          Creates a new task within a collaboration. If no `organization_ids`
          are given the task is send to all organizations within the
          collaboration. The endpoint can be accessed by both a `User` and
          `Container`.\n

          ### Permission Table\n
          |Rule name|Scope|Operation|Assigned to node|Assigned to container|
          Description|\n
          |--|--|--|--|--|--|\n
          |Task|Global|Create|❌|❌|Create a new task|\n
          |Task|Collaboration|Create|❌|✅|Create a new task for a specific
          collaboration in which your organization participates|\n

          ## Accessed as `User`\n
          This endpoint is accessible to users. A new `job_id` is
          created when a user creates a task. The user needs to be within an
          organization that is part of the collaboration to which the task is
          posted.\n

          ## Accessed as `Container`\n
          When this endpoint is accessed by an algorithm container, it is
          considered to be a child-task of the container, and will get the
          `job_id` from the initial task. Containers have limited permissions
          to create tasks: they are only allowed to create tasks in the same
          collaboration using the same image.\n

        requestBody:
          content:
            application/json:
              schema:
                $ref: '#/components/schemas/Task'

        responses:
          200:
            description: Ok
          400:
            description: Wrong input, or not all required nodes are registered, or you
              are not in the collaboration yourself
          401:
            description: Unauthorized
          403:
            description: Algorithm store is not part of the collaboration
          404:
            description: Collaboration with `collaboration_id` not found

        security:
          - bearerAuth: []

        tags: ["Task"]
        """
        try:
            request.get_json()
        except Exception:
            return {"msg": "Request body is incorrect"}, HTTPStatus.BAD_REQUEST
        return self.post_task(request.get_json(), self.socketio, self.r, self.config)

    # TODO this function should be refactored to make it more readable
    @staticmethod
    def post_task(data: dict, socketio: SocketIO, rules: RuleCollection, config: dict):
        """
        Create new task and algorithm runs. Send the task to the nodes.

        Parameters
        ----------
        data : dict
            Task data
        socketio : SocketIO
            SocketIO server instance
        rules : RuleCollection
            Rule collection instance
        config : dict
            Configuration dictionary
        """
        # validate request body
        errors = task_input_schema.validate(data)
        if errors:
            return {
                "msg": "Request body is incorrect",
                "errors": errors,
            }, HTTPStatus.BAD_REQUEST

        # A task can be created for a collaboration or a study. If it is for a study,
        # a study_id is always given, and a collaboration_id is optional. If it is for
        # a collaboration, a collaboration_id is always given, and a study_id is
        # never set. The following logic checks if the given study_id and
        # collaboration_id are valid and when both are provided, checks if they match.
        collaboration_id = data.get("collaboration_id")
        study_id = data.get("study_id")
        study = None
        if collaboration_id:
            collaboration = db.Collaboration.get(collaboration_id)
            if not collaboration:
                return {
                    "msg": f"Collaboration id={collaboration_id} not found!"
                }, HTTPStatus.NOT_FOUND
        if study_id:
            study = db.Study.get(study_id)
            if not study:
                return {"msg": f"Study id={study_id} not found"}, HTTPStatus.NOT_FOUND

            # check if collaboration and study match if both are set
            if collaboration_id and study.collaboration_id != collaboration_id:
                return {
                    "msg": (
                        f"The study_id '{study.id}' does not belong to the "
                        f"collaboration_id '{collaboration_id}' that is given."
                    )
                }, HTTPStatus.BAD_REQUEST

            # get the collaboration object as well
            collaboration_id = study.collaboration_id
            collaboration = db.Collaboration.get(collaboration_id)

        organizations_json_list = data.get("organizations")
        org_ids = [org.get("id") for org in organizations_json_list]
        db_ids = collaboration.get_organization_ids()

        # Check that all organization ids are within the collaboration, this
        # also ensures us that the organizations exist
        if not set(org_ids).issubset(db_ids):
            return {
                "msg": (
                    "At least one of the supplied organizations in not within "
                    "the collaboration."
                )
            }, HTTPStatus.BAD_REQUEST
        # check that they are within the study (if that has been defined)
        if study:
            study_org_ids = [org.id for org in study.organizations]
            if not set(org_ids).issubset(study_org_ids):
                return {
                    "msg": (
                        "At least one of the supplied organizations in not within "
                        "the specified study."
                    )
                }, HTTPStatus.BAD_REQUEST

        # check if all the organizations have a registered node
        nodes = (
            g.session.query(db.Node)
            .filter(db.Node.organization_id.in_(org_ids))
            .filter(db.Node.collaboration_id == collaboration_id)
            .all()
        )
        if len(nodes) < len(org_ids):
            present_nodes = [node.organization_id for node in nodes]
            missing = [str(id) for id in org_ids if id not in present_nodes]
            return {
                "msg": (
                    "Cannot create this task because there are no nodes registered"
                    f" for the following organization(s): {', '.join(missing)}."
                )
            }, HTTPStatus.BAD_REQUEST
        # check if any of the nodes that are offline shared their configuration
        # info and if this prevents this user from creating this task
        if g.user:
            for node in nodes:
                if Tasks._node_doesnt_allow_user_task(node.config):
                    return {
                        "msg": (
                            "Cannot create this task because one of the nodes that"
                            " you are trying to send this task to does not allow "
                            "you to create tasks."
                        )
                    }, HTTPStatus.BAD_REQUEST

        # figure out the initiating organization of the task
        if g.user:
            init_org = g.user.organization
        else:  # g.container:
            init_org = db.Node.get(g.container["node_id"]).organization

        # check if the initiating organization is part of the collaboration
        if init_org not in collaboration.organizations:
            return {
                "msg": "You can only create tasks for collaborations "
                "you are participating in!"
            }, HTTPStatus.UNAUTHORIZED

        # verify permissions
        image = data.get("image", "")
        if g.user and not rules.can_for_col(P.CREATE, collaboration.id):
            return {
                "msg": "You lack the permission to do that!"
            }, HTTPStatus.UNAUTHORIZED

        elif g.container:
            # verify that the container has permissions to create the task
            if not Tasks.__verify_container_permissions(
                g.container, image, collaboration_id
            ):
                return {"msg": "Container-token is not valid"}, HTTPStatus.UNAUTHORIZED

        # get the algorithm store
        if g.user:
            store_id = data.get("store_id")
            store = None
            if store_id:
                store = db.AlgorithmStore.get(store_id)
                if not store:
                    return {
                        "msg": f"Algorithm store id={store_id} not found!"
                    }, HTTPStatus.BAD_REQUEST
                # check if the store is part of the collaboration
                if (
                    not store.is_for_all_collaborations()
                    and store.collaboration_id != collaboration_id
                ):
                    return {
                        "msg": (
                            "The algorithm store is not part of the collaboration "
                            "to which the task is posted."
                        )
                    }, HTTPStatus.FORBIDDEN
                # get the algorithm from the algorithm store
                try:
                    image, digest = Tasks._get_image_and_hash_from_store(
                        store=store,
                        image=image,
                        config=config,
                        server_url_from_request=data.get("server_url"),
                    )
                except Exception as e:
                    log.exception("Error while getting image from store: %s", e)
                    return {"msg": str(e)}, HTTPStatus.BAD_REQUEST

                if digest:
                    image_with_hash = f"{image}@{digest}"
                else:
                    # hash lookup in store was unsuccessful, use image without hash, but
                    # also set store to None as it was not successfully looked up
                    image_with_hash = image
                    store = None
            else:
                # no need to determine hash if we don't look it up in a store
                image_with_hash = image
        else:  # this is a container task
            parent = db.Task.get(g.container["task_id"])
            store = parent.algorithm_store
            image_with_hash = parent.image

        # check that the input is valid. If the collaboration is encrypted, it
        # should not be possible to read the input, and we should not save it
        # to the database as it may be sensitive information. Vice versa, if
        # the collaboration is not encrypted, we should not allow the user to
        # send encrypted input.
        blob_storage_used = bool(config.get("large_result_store", {}))

        is_valid_input, error_msg = Tasks._check_input(
            organizations_json_list, collaboration, blob_storage_used
        )
        if not is_valid_input:
            return {"msg": error_msg}, HTTPStatus.BAD_REQUEST

        # permissions ok, create task record and TaskDatabase records
        task = db.Task(
            collaboration=collaboration,
            study=study,
            name=data.get("name", ""),
            description=data.get("description", ""),
            image=image_with_hash,
            init_org=init_org,
            algorithm_store=store,
            created_at=datetime.datetime.now(datetime.timezone.utc),
        )

        # create job_id. Users can only create top-level -tasks (they will not
        # have sub-tasks). Therefore, always create a new job_id. Tasks created
        # by containers are always sub-tasks
        if g.user:
            task.job_id = task.next_job_id()
            task.init_user_id = g.user.id
            log.debug(f"New job_id {task.job_id}")
        elif g.container:
            task.parent_id = g.container["task_id"]
            parent = db.Task.get(g.container["task_id"])
            task.job_id = parent.job_id
            task.init_user_id = parent.init_user_id
            log.debug(f"Sub task from parent_id={task.parent_id}")

        # save the databases that the task uses
        databases = data.get("databases")
        if isinstance(databases, str):
            databases = [{"label": databases}]
        elif databases is None:
            databases = []
        db_records = []
        for database in databases:
            if "label" not in database:
                return {
                    "msg": "Database label missing! The dictionary "
                    f"{database} should contain a 'label' key"
                }, HTTPStatus.BAD_REQUEST
            # remove label from the database dictionary, which apart from it
            # may only contain some optional parameters . Save optional
            # parameters as JSON without spaces to database
            label = database.pop("label")
            # TODO task.id is only set here because in between creating the
            # task and using the ID here, there are other database operations
            # that silently update the task.id (i.e. next_job_id() and
            # db.Task.get()). Task.id should be updated explicitly instead.
            db_records.append(
                db.TaskDatabase(
                    task_id=task.id,
                    database=label,
                    parameters=json.dumps(database, separators=(",", ":")),
                )
            )

        # All checks completed, save task to database
        task.save()
        [db_record.save() for db_record in db_records]  # pylint: disable=W0106

        # now we need to create results for the nodes to fill. Each node
        # receives their instructions from a result, not from the task itself
        log.debug(f"Assigning task to {len(organizations_json_list)} nodes.")
        for org in organizations_json_list:
            organization = db.Organization.get(org["id"])
            log.debug(f"Assigning task to '{organization.name}'.")
            input_ = org.get("input")
            # FIXME: legacy input from the client, could be removed at some
            # point
            if isinstance(input_, dict):
                input_ = json.dumps(input_).encode(STRING_ENCODING)
            # Create run
            run = db.Run(
                task=task,
                organization=organization,
                input=input_,
                status=TaskStatus.PENDING,
                blob_storage_used=blob_storage_used,
            )
            run.save()

        # notify nodes a new task available (only to online nodes), nodes that
        # are offline will receive this task on sign in.
        socketio.emit(
            "new_task",
            {"id": task.id, "parent_id": task.parent_id},
            namespace="/tasks",
            room=f"collaboration_{task.collaboration_id}",
        )

        # add some logging
        log.info(f"New task for collaboration '{task.collaboration.name}'")
        if g.user:
            log.debug(f" created by: '{g.user.username}'")
        else:
            log.debug(
                f" created by container on node_id="
                f"{g.container['node_id']}"
                f" for (parent) task_id={g.container['task_id']}"
            )

        log.debug(f" url: '{url_for('task_with_id', id=task.id)}'")
        log.debug(f" name: '{task.name}'")
        log.debug(f" image: '{task.image}'")

        return task_schema.dump(task, many=False), HTTPStatus.CREATED

    @staticmethod
    def __verify_container_permissions(container, image, collaboration_id):
        """Validates that the container is allowed to create the task."""

        # check that the image is allowed: algorithm containers can only
        # create tasks with the same image
        if not image.endswith(container["image"]):
            log.warning(
                (
                    f"Container from node={container['node_id']} "
                    f"attempts to post a task using illegal image!?"
                )
            )
            log.warning(f"  task image: {image}")
            log.warning(f"  container image: {container['image']}")
            return False

        # check that parent task is not completed yet
        if has_task_finished(db.Task.get(container["task_id"]).status):
            log.warning(
                f"Container from node={container['node_id']} "
                f"attempts to start sub-task for a completed "
                f"task={container['task_id']}"
            )
            return False

        # check that node id is indeed part of the collaboration
        if not container["collaboration_id"] == collaboration_id:
            log.warning(
                f"Container attempts to create a task outside "
                f"collaboration_id={container['collaboration_id']} in "
                f"collaboration_id={collaboration_id}!"
            )
            return False

        return True

    @staticmethod
    def _node_doesnt_allow_user_task(node_configs: list[db.NodeConfig]) -> bool:
        """
        Checks if the node allows user to create task.

        Parameters
        ----------
        node_configs : list[db.NodeConfig]
            List of node configurations.

        Returns
        -------
        bool
            True if the node doesn't allow the user to create task.
        """
        has_limitations = False
        for config_option in node_configs:
            if config_option.key == NodePolicy.ALLOWED_USERS:
                has_limitations = True
                # TODO expand when we allow also usernames, like orgs below
                if g.user.id == int(config_option.value):
                    return False
            elif config_option.key == "allowed_orgs":
                has_limitations = True
                if config_option.value.isdigit():
                    if g.user.organization_id == int(config_option.value):
                        return False
                else:
                    org = db.Organization.get_by_name(config_option.value)
                    if org and g.user.organization_id == org.id:
                        return False
        return has_limitations

    @staticmethod
    def _check_input(
        organizations_json_list: list[dict],
        collaboration: db.Collaboration,
        blob_storage_used: bool,
    ) -> tuple[bool, str]:
        """
        Check if the input is valid for the collaboration. If the collaboration
        is encrypted, the input should not be readable as a string. If the
        collaboration is not encrypted, the input should be readable as a string.

        Parameters
        ----------
        organizations_json_list : list[dict]
            List of organizations which contains the input per organization.
        collaboration : db.Collaboration
            Collaboration object.
        blob_storage_used : bool
            Whether or not blob storage is used for storing data.

        Returns
        -------
        bool
            True if the input is valid.
        str
            Error message if the input is invalid.
        """
        if not organizations_json_list:
            return False, "No organizations provided in the request."
        if blob_storage_used:
            return Tasks.check_input_uuid(organizations_json_list)
        else:
            return Tasks._check_input_encryption(organizations_json_list, collaboration)

    @staticmethod
    def check_input_uuid(organizations_json_list: list[dict]) -> tuple[bool, str]:
        """
        Check if the input is a valid uuid for all collaborations.
        Parameters
        ----------
        organizations_json_list : list[dict]
            List of organizations containing the uuids of the inputs.

        Returns
        -------
        bool
            True if the input is valid.
        str
            Error message if the input is invalid.
        """
        for org in organizations_json_list:
            input_ = org.get("input")
            if not isinstance(input_, str):
                return False, (
                    "Your task's input cannot be parsed. Your input should be "
                    "a UUID as a large result store has been configured. Note that if you are using "
                    "the user interface or Python client, this should be done "
                    "for you."
                )
            try:
                uuid.UUID(input_)
            except ValueError:
                return False, (
                    "Your task's input cannot be parsed. Note that if you are using "
                    "the user interface or Python client, this should be done "
                    "for you."
                )
        return True, ""

    @staticmethod
    def _check_input_encryption(
        organizations_json_list: list[dict], collaboration: db.Collaboration
    ) -> tuple[bool, str]:
        """
        Check if the input encryption status matches the expected status for
        the collaboration. Also, check that if the input is not encrypted, it
        can be read as a string.

        Parameters
        ----------
        organizations_json_list : list[dict]
            List of organizations which contains the input per organization.
        collaboration : db.Collaboration
            Collaboration object.

        Returns
        -------
        bool
            True if the input is encrypted.
        str
            Error message if the input is valid.
        """
        dummy_cryptor = DummyCryptor()
        for org in organizations_json_list:
            input_ = org.get("input")
            decrypted_input = dummy_cryptor.decrypt(input_)
            is_input_readable = False
            try:
                decrypted_input.decode(STRING_ENCODING)
                is_input_readable = True
            except UnicodeDecodeError:
                pass

            if collaboration.encrypted and is_input_readable:
                return (
                    False,
                    (
                        "Your collaboration requires encryption, but input is not "
                        "encrypted! Please encrypt your input before sending it."
                    ),
                )
            elif not collaboration.encrypted and not is_input_readable:
                return False, (
                    "Your task's input cannot be parsed. Your input should be "
                    "a base64 encoded JSON string. Note that if you are using "
                    "the user interface or Python client, this should be done "
                    "for you. Also, make sure not to encrypt your input, "
                    "as your collaboration is set to not use encryption."
                )
        return True, ""

    @staticmethod
    def _get_image_and_hash_from_store(
        store: db.AlgorithmStore,
        image: str,
        config: dict,
        server_url_from_request: str | None = None,
    ) -> tuple[str, str]:
        """
        Determine the image and hash from the algorithm store.

        Parameters
        ----------
        store : db.AlgorithmStore
            Algorithm store.
        image : str
            URL of the docker image to be used.
        config : dict
            Configuration dictionary.
        server_url_from_request : str, optional
            Server URL from the request, by default None

        Returns
        -------
        tuple[str, str]
            Image url and image hash digest.

        Raises
        ------
        Exception
            If the algorithm cannot be retrieved from the store.
        """
        server_url = get_server_url(config, server_url_from_request)
        if not server_url:
            raise ValueError(
                "Server URL is not set in the configuration nor in the request "
                "arguments. Please provide it as 'server_url' in the request."
            )
        # get the algorithm from the store
        response, status_code = request_algo_store(
            algo_store_url=store.url,
            server_url=server_url,
            endpoint="algorithm",
            method="GET",
            params={"image": image},
        )
        if status_code != HTTPStatus.OK:
            raise Exception(
                f"Could not retrieve algorithm from store! {response.get('msg')}"
            )
        try:
            algorithm = response.json()["data"][0]
        except Exception as e:
            raise Exception("Algorithm not found in store!") from e

        image = algorithm["image"]
        digest = algorithm["digest"]
        # TODO v5+ remove this check? The digest should always be present for an
        # algorithm from stores started in v4.6 and higher
        if image and not digest:
            log.warning(
                "Algorithm image %s does not have a digest in the algorithm store. Will"
                " use it without digest.",
                image,
            )
            return image, None
        return image, digest


class Task(TaskBase):
    """Resource for /api/task"""

    @only_for(("user", "node", "container"))
    def get(self, id):
        """Get task
        ---
        description: >-
          Returns the task specified by the id.

          ### Permission Table\n
          |Rule name|Scope|Operation|Assigned to node|Assigned to container|
          Description|\n
          |--|--|--|--|--|--|\n
          |Task|Global|View|❌|❌|View any task|\n
          |Task|Collaboration|View|✅|✅|View any task in your collaborations|
          |Task|Organization|View|❌|❌|View any task that your organization
          created|\n
          |Task|Own|View|❌|❌|View any task that you created|\n

          Accessible to users.

        parameters:
          - in: path
            name: id
            schema:
              type: integer
            description: Task id
            required: true
          - in: query
            name: include
            schema:
              type: string
            description: Include 'results' to include the task's results,
              'runs' to include details on algorithm runs. For including
              multiple, do either `include=x,y` or `include=x&include=y`.

        responses:
          200:
            description: Ok
          404:
            description: Task not found
          401:
            description: Unauthorized

        security:
          - bearerAuth: []

        tags: ["Task"]
        """
        task = db.Task.get(id)
        if not task:
            return {"msg": f"task id={id} is not found"}, HTTPStatus.NOT_FOUND

        # obtain schema
        schema = self._select_schema()

        # check permissions
        if not self.r.allowed_for_org(P.VIEW, task.init_org_id) and not (
            self.r.v_own.can() and g.user and task.init_user_id == g.user.id
        ):
            return {
                "msg": "You lack the permission to do that!"
            }, HTTPStatus.UNAUTHORIZED
        # if results are included, check permissions for results
        if (
            self.is_included("results")
            and not self.r_run.allowed_for_org(P.VIEW, task.init_org_id)
            and not (self.r.v_own.can() and g.user and task.init_user_id == g.user.id)
        ):
            return {
                "msg": "You lack the permission to view results for this " "task!"
            }, HTTPStatus.UNAUTHORIZED

        return schema.dump(task, many=False), HTTPStatus.OK

    @with_user
    def delete(self, id):
        """Remove task
        ---
        description: >-
          Remove tasks and their runs.\n

          ### Permission Table\n
          |Rule name|Scope|Operation|Assigned to node|Assigned to container|
          Description|\n
          |--|--|--|--|--|--|\n
          |Task|Global|Delete|❌|❌|Delete a task|\n
          |Task|Collaboration|Delete|❌|❌|Delete a task from a collaboration
          in which your organization participates|\n
          |Task|Organization|Delete|❌|❌|Delete a task that your organization
          initiated|\n
          |Task|Own|Delete|❌|❌|Delete a task you created yourself|\n

          Accessible to users.

        parameters:
          - in: path
            name: id
            schema:
              type: integer
            description: Task id
            required: true

        responses:
          200:
            description: Ok
          404:
            description: Task not found
          401:
            description: Unauthorized

        security:
          - bearerAuth: []

        tags: ["Task"]
        """

        task = db.Task.get(id)
        if not task:
            return {"msg": f"Task id={id} not found"}, HTTPStatus.NOT_FOUND

        # validate permissions
        if not self.r.allowed_for_org(P.DELETE, task.init_org_id) and not (
            self.r.d_own.can() and task.init_user_id == g.user.id
        ):
            return {
                "msg": "You lack the permission to do that!"
            }, HTTPStatus.UNAUTHORIZED

        # kill the task if it is still running
        if not has_task_finished(task.status):
            kill_task(task, self.socketio)

        # retrieve runs that belong to this task
        log.info(f"Removing task id={task.id}")
        for run in task.runs:
            log.info(f" Removing run id={run.id}")
            run.delete()

        # delete child/grandchild/... tasks
        Task._delete_subtasks(task)
        # permissions ok, delete...
        task.delete()

        return {
            "msg": f"task id={id} and its algorithm run data have been "
            "successfully deleted"
        }, HTTPStatus.OK

    @staticmethod
    def _delete_subtasks(task: db.Task) -> None:
        """
        Delete subtasks recursively.

        Parameters
        ----------
        task : db.Task
            Task to delete.
        """
        for child_task in task.children:
            Task._delete_subtasks(child_task)
            log.info(f" Removing child task id={child_task.id}")
            child_task.delete()


class TaskStatusEndpoint(TaskBase):
    """Resource for /api/task/<id>/status"""

    @only_for(("user", "container"))
    def get(self, task_id: int):
        """Get task status
        ---
        description: >-
          Returns the status of the task specified by the id.

          ### Permission Table\n
          |Rule name|Scope|Operation|Assigned to node|Assigned to container|
          Description|\n
          |--|--|--|--|--|--|\n
          |Task|Global|View|❌|❌|View any task|\n
          |Task|Collaboration|View|✅|✅|View any task in your collaborations|
          |Task|Organization|View|❌|❌|View any task that your organization
          created|\n
          |Task|Own|View|❌|❌|View any task that you created|\n

          Accessible to users.

        parameters:
          - in: path
            name: id
            schema:
              type: integer
            description: Task id
            required: true

        responses:
          200:
            description: Ok
            content:
              application/json:
                schema:
                  type: object
                  properties:
                    status:
                      type: string
                      description: The status of the task
          404:
            description: Task not found
          401:
            description: Unauthorized

        security:
          - bearerAuth: []

        tags: ["Task"]
        """
        task = db.Task.get(task_id)
        if not task:
            log.error(f"Task with id={task_id} not found.")
            return {"msg": f"Task id={task_id} not found"}, HTTPStatus.NOT_FOUND

        if not self._has_permission_to_view_task(task):
            log.error(f"Unauthorized access to task id={task_id}.")
            return {
                "msg": "You lack the permission to do that!"
            }, HTTPStatus.UNAUTHORIZED

        log.info(f"Returning status for task id={task_id}: {task.status}")
        return {"status": task.status}, HTTPStatus.OK

    def _has_permission_to_view_task(self, task: db.Task) -> bool:
        """
        Check if the user has permission to view the task.

        Parameters
        ----------
        task : db.Task
            Task to check permissions for.

        Returns
        -------
        bool
            True if the user has permission, False otherwise.
        """
        return self.r.allowed_for_org(P.VIEW, task.init_org_id) or (
            self.r.v_own.can() and g.user and task.init_user_id == g.user.id
        )



"""
The server has a central function in the vantage6 architecture. It stores
in the database which organizations, collaborations, users, etc.
exist. It allows the users and nodes to authenticate and subsequently interact
through the API the server hosts. Finally, it also communicates with
authenticated nodes and users via the socketIO server that is run here.
"""

import os
from gevent import monkey

# This is a workaround for readthedocs
if not os.environ.get("READTHEDOCS"):
    # flake8: noqa: E402 (ignore import error)
    monkey.patch_all()

# pylint: disable=wrong-import-position, wrong-import-order
import enum
import importlib
import logging
import uuid
import json
import time
import datetime as dt
import traceback
from requests import Response as RequestsResponse


from http import HTTPStatus
from werkzeug.exceptions import HTTPException

from flask import (
    Flask,
    make_response,
    current_app,
    request,
    send_from_directory,
    Request,
    Response,
)
from flask_cors import CORS
from flask_cors.core import probably_regex
from flask_jwt_extended import JWTManager
from flask_marshmallow import Marshmallow
from flask_restful import Api
from flask_principal import Principal, Identity, identity_changed
from flask_socketio import SocketIO
from threading import Thread
from pathlib import Path
from sqlalchemy.orm.exc import NoResultFound

from vantage6.common import logger_name, split_rabbitmq_uri
from vantage6.common.globals import (
    PING_INTERVAL_SECONDS,
    AuthStatus,
    DEFAULT_PROMETHEUS_EXPORTER_PORT,
)
from vantage6.backend.common.globals import HOST_URI_ENV, DEFAULT_SUPPORT_EMAIL_ADDRESS
from vantage6.backend.common.jsonable import jsonable
from vantage6.backend.common.permission import RuleNeed
from vantage6.backend.common.metrics import Metrics, start_prometheus_exporter
from vantage6.backend.common.mail_service import MailService
from vantage6.cli.context.server import ServerContext
from vantage6.server.model.base import DatabaseSessionManager, Database
from vantage6.server.permission import PermissionManager
from vantage6.server import db
from vantage6.server.resource.common.output_schema import HATEOASModelSchema
from vantage6.server.globals import (
    APPNAME,
    ACCESS_TOKEN_EXPIRES_HOURS,
    RESOURCES,
    RESOURCES_PATH,
    SUPER_USER_INFO,
    REFRESH_TOKENS_EXPIRE_HOURS,
    MIN_TOKEN_VALIDITY_SECONDS,
    MIN_REFRESH_TOKEN_EXPIRY_DELTA,
    SERVER_MODULE_NAME,
)
from vantage6.server.websockets import DefaultSocketNamespace
from vantage6.server.default_roles import get_default_roles, DefaultRole
from vantage6.server.hashedpassword import HashedPassword
from vantage6.server.controller import cleanup
from vantage6.server.service.azure_storage_service import AzureStorageService


# make sure the version is available
from vantage6.server._version import __version__  # noqa: F401


module_name = logger_name(__name__)
log = logging.getLogger(module_name)


class ServerApp:
    """
    Vantage6 server instance.

    Attributes
    ----------
    ctx : ServerContext
        Context object that contains the configuration of the server.
    """

    def __init__(self, ctx: ServerContext) -> None:
        """Create a vantage6-server application."""

        self.ctx = ctx

        # initialize, configure Flask
        self.app = Flask(
            SERVER_MODULE_NAME,
            root_path=Path(__file__),
            template_folder=Path(__file__).parent / "templates",
            static_folder=Path(__file__).parent / "static",
        )
        self.debug: dict = self.ctx.config.get("debug", {})
        self.configure_flask()

        # Setup SQLAlchemy and Marshmallow for marshalling/serializing
        self.ma = Marshmallow(self.app)

        # Setup the Flask-JWT-Extended extension (JWT: JSON Web Token)
        self.jwt = JWTManager(self.app)
        self.configure_jwt()

        # Setup Principal, granular API access manegement
        self.principal = Principal(self.app, use_sessions=False)

        # Enable cross-origin resource sharing. Note that Flask-CORS interprets
        # the origins as regular expressions.
        cors_allowed_origins = self.ctx.config.get("cors_allowed_origins", "*")
        self._warn_if_cors_regex(cors_allowed_origins)
        self.cors = CORS(
            self.app,
            resources={r"/*": {"origins": cors_allowed_origins}},
        )

        # Setup the Flask-Mail client
        self.mail = MailService(self.app)

        self.metrics = Metrics(labels=["node_id", "platform", "os"])
        # Setup websocket channel
        self.socketio = self.setup_socket_connection()

        self.setup_large_result_store()
        # setup the permission manager for the API endpoints
        self.permissions = PermissionManager(RESOURCES_PATH, RESOURCES, DefaultRole)

        # Api - REST JSON-rpc
        self.api = Api(self.app)
        self.configure_api()
        self.load_resources()

        # set environment variable for dev environment
        host_uri = self.ctx.config.get("dev", {}).get("host_uri")
        if host_uri:
            os.environ[HOST_URI_ENV] = host_uri

        # couple any algoritm stores to the server if defined in config. This should be
        # done after the resources are loaded to ensure that rules are set up
        # self.couple_algorithm_stores()

        # TODO v5+ clean this up (simply delete community store URL update). This
        # change is here to prevent errors when going from v4.3-4.7 to 4.8+.
        # Because in v4.8, the algorithm store's API path was made
        # flexible, the /api from then on had to be included in the database
        community_stores = db.AlgorithmStore.get_by_url(
            "https://store.cotopaxi.vantage6.ai"
        )
        if community_stores:
            # only need to update first one (there shouldn't be more than 1 anyway)
            community_store = community_stores[0]
            log.warning(
                "Updating community store URL to include '/api'. This change is"
                " necessary as you are updating to v4.8"
            )
            community_store.url = f"{community_store.url}/api"
            community_store.save()

        # set the server version
        self.__version__ = __version__

        if self.ctx.config.get("runs_data_cleanup_days"):
            log.info(
                "Results older than %s days will be removed",
                self.ctx.config.get("runs_data_cleanup_days"),
            )
            t_cleanup = Thread(target=self.__runs_data_cleanup_worker, daemon=True)
            t_cleanup.start()

        # set up socket ping/pong
        log.debug("Starting thread to set node status")
        t = Thread(target=self.__node_status_worker, daemon=True)
        t.start()

        if self.ctx.config.get("prometheus", {}).get("enabled", False):
            start_prometheus_exporter(
                port=self.ctx.config.get(
                    "prometheus_port", DEFAULT_PROMETHEUS_EXPORTER_PORT
                )
            )

        log.info("Initialization done")

    def setup_large_result_store(self):
        """
        Setup the large result store for storing large results.
        If configured, inputs and results will be stored in blob Storage.
        """

        self.storage_adapter = None
        large_result_config = self.ctx.config.get("large_result_store", {})
        if not large_result_config:
            log.info(
                "No large result store configured, using relational database for input and result storage"
            )
            return

        log.info("Using Azure Blob Storage as large result store")
        self.storage_adapter = AzureStorageService(config=large_result_config)

    @staticmethod
    def _warn_if_cors_regex(origins: str | list[str]) -> None:
        """
        Give a warning if CORS origins are regular expressions. This will not work
        properly for socket events (Flask-SocketIO checks for string equality and does
        not use regex).

        Note that we are using the `probably_regex` function from Flask-CORS to check
        if the origins are probably regular expressions - the Flask implementation for
        determining if it is a regex is a bit hacky (see
        https://github.com/corydolphin/flask-cors/blob/3.0.10/flask_cors/core.py#L275-L285)
        and Flask-CORS doesn't currently offer an opt out of regex's altogether.

        Parameters
        ----------
        origins: str | list[str]
            The origins to check
        """
        if isinstance(origins, str):
            origins = [origins]

        for origin in origins:
            if probably_regex(origin) and not origin == "*":
                log.warning(
                    "CORS origin '%s' is a regular expression. Socket events sent from "
                    "this origin will not be handled properly.",
                    origin,
                )

    def setup_socket_connection(self) -> SocketIO:
        """
        Setup a socket connection. If a message queue is defined, connect the
        socket to the message queue. Otherwise, use the default socketio
        settings.

        Returns
        -------
        SocketIO
            SocketIO object
        """
        msg_queue = self.ctx.config.get("rabbitmq", {}).get("uri")
        if msg_queue:
            try:
                splitted_rabbit_uri = split_rabbitmq_uri(msg_queue)
                log.debug(
                    "Connecting to msg queue: amqp://<user>:<pass>@%s:%s/%s",
                    splitted_rabbit_uri["host"],
                    splitted_rabbit_uri["port"],
                    splitted_rabbit_uri["vhost"],
                )
            except Exception:
                log.warning(
                    "Failed to parse RabbitMQ URI. Will try to use the provided"
                    "URI to connect, but it may likely fail."
                )

        debug_mode = self.debug.get("socketio", False)
        if debug_mode:
            log.debug("SocketIO debug mode enabled")

        cors_settings = self.ctx.config.get("cors_allowed_origins", "*")
        try:
            socketio = SocketIO(
                self.app,
                async_mode="gevent_uwsgi",
                message_queue=msg_queue,
                ping_timeout=60,
                cors_allowed_origins=cors_settings,
                logger=debug_mode,
                engineio_logger=debug_mode,
                always_connect=True,
            )
        except Exception as e:
            log.warning(
                "Default socketio settings failed, attempt to run "
                "without gevent_uwsgi packages! This leads to "
                "performance issues and possible issues concerning "
                "the websocket channels!"
            )
            log.debug(e)
            socketio = SocketIO(
                self.app,
                message_queue=msg_queue,
                ping_timeout=60,
                cors_allowed_origins=cors_settings,
                logger=debug_mode,
                engineio_logger=debug_mode,
                always_connect=True,
            )

        namespace = DefaultSocketNamespace("/tasks", socketio, self.metrics)
        socketio.on_namespace(namespace)

        return socketio

    def configure_flask(self) -> None:
        """Configure the Flask settings of the vantage6 server."""

        # let us handle exceptions
        self.app.config["PROPAGATE_EXCEPTIONS"] = True

        # patch where to obtain token
        self.app.config["JWT_AUTH_URL_RULE"] = "/api/token"

        # PyJWT 2.10+ enforces that `sub` is a string. Vantage6 historically
        # used non-string subjects (e.g. ints and container dicts), which
        # breaks protected endpoints after the PyJWT bump.
        # Disabling this validation keeps compatibility
        self.app.config["JWT_VERIFY_SUB"] = False

        # If no secret is set in the config file, one is generated. This
        # implies that all (even refresh) tokens will be invalidated on restart
        self.app.config["JWT_SECRET_KEY"] = self.ctx.config.get(
            "jwt_secret_key", str(uuid.uuid4())
        )

        # Default expiration time
        token_expiry_seconds = self._get_jwt_expiration_seconds(
            config_key="token_expires_hours", default_hours=ACCESS_TOKEN_EXPIRES_HOURS
        )
        self.app.config["JWT_ACCESS_TOKEN_EXPIRES"] = token_expiry_seconds

        # Set refresh token expiration time
        self.app.config["JWT_REFRESH_TOKEN_EXPIRES"] = self._get_jwt_expiration_seconds(
            config_key="refresh_token_expires_hours",
            default_hours=REFRESH_TOKENS_EXPIRE_HOURS,
            longer_than=token_expiry_seconds + MIN_REFRESH_TOKEN_EXPIRY_DELTA,
            is_refresh=True,
        )

        # # Open Api Specification (f.k.a. swagger)
        # self.app.config["SWAGGER"] = {
        #     "title": APPNAME,
        #     "uiversion": "3",
        #     "openapi": "3.0.0",
        #     "version": __version__,
        # }

        # Mail settings
        mail_config = self.ctx.config.get("smtp", {})
        self.app.config["MAIL_PORT"] = mail_config.get("port", 1025)
        self.app.config["MAIL_SERVER"] = mail_config.get("server", "localhost")
        self.app.config["MAIL_USERNAME"] = mail_config.get(
            "username", DEFAULT_SUPPORT_EMAIL_ADDRESS
        )
        self.app.config["MAIL_PASSWORD"] = mail_config.get("password", "")
        self.app.config["MAIL_USE_TLS"] = mail_config.get("MAIL_USE_TLS", True)
        self.app.config["MAIL_USE_SSL"] = mail_config.get("MAIL_USE_SSL", False)
        debug_mode = self.debug.get("flask", False)
        if debug_mode:
            log.debug("Flask debug mode enabled")
        self.app.debug = debug_mode

        def _get_request_path(request: Request) -> str:
            """
            Return request extension of request URL, e.g.
            http://localhost:7601/api/task/1 -> api/task/1

            Parameters
            ----------
            request: Request
                Flask request object

            Returns
            -------
            string:
                The endpoint path of the request
            """
            return request.url.replace(request.url_root, "")

        # before request
        @self.app.before_request
        def do_before_request():
            """Before every flask request method."""
            # Add log message before each request
            log.debug(
                f"Received request: {request.method} {_get_request_path(request)}"
            )

            # This will obtain a (scoped) db session from the session factory
            # that is linked to the flask request global `g`. In every endpoint
            # we then can access the database by using this session. We ensure
            # that the session is removed (and uncommited changes are rolled
            # back) at the end of every request.
            DatabaseSessionManager.new_session()

        @self.app.after_request
        def remove_db_session(response):
            """After every flask request.

            This will close the database session created by the
            `before_request`.
            """
            DatabaseSessionManager.clear_session()
            return response

        @self.app.errorhandler(HTTPException)
        def error_remove_db_session(error: HTTPException):
            """In case an HTTP-exception occurs during the request.

            It is important to close the db session to avoid having dangling
            sessions.
            """
            if error.code == 404:
                log.debug("404 error for route '%s'", _get_request_path(request))
            else:
                log.warning("HTTP Exception occured during request")
                log.debug("Details exception: %s", traceback.format_exc())
            DatabaseSessionManager.clear_session()
            return error.get_response()

        @self.app.errorhandler(Exception)
        def error2_remove_db_session(error):
            """In case an exception occurs during the request.

            It is important to close the db session to avoid having dangling
            sessions.
            """
            log.exception("Exception occured during request")
            DatabaseSessionManager.clear_session()
            return {
                "msg": "An unexpected error occurred on the server!"
            }, HTTPStatus.INTERNAL_SERVER_ERROR

        @self.app.route("/robots.txt")
        def static_from_root():
            return send_from_directory(self.app.static_folder, request.path[1:])

    def _get_jwt_expiration_seconds(
        self,
        config_key: str,
        default_hours: int,
        longer_than: int = MIN_TOKEN_VALIDITY_SECONDS,
        is_refresh: bool = False,
    ) -> int:
        """
        Return the expiration time for JWT tokens.

        This time may be specified in the config file. If it is not, the
        default value is returned.

        Parameters
        ----------
        config_key: str
            The config key to look for that sets the expiration time
        default_hours: int
            The default expiration time in hours
        longer_than: int
            The minimum expiration time in hours.
        is_refresh: bool
            If True, the expiration time is for a refresh token. If False, it
            is for an access token.

        Returns
        -------
        int:
            The JWT token expiration time in seconds
        """
        hours_expire = self.ctx.config.get(config_key)
        if hours_expire is None:
            # No value is present in the config file, use default
            seconds_expire = int(float(default_hours) * 3600)
        elif (
            isinstance(hours_expire, (int, float))
            or hours_expire.replace(".", "").isnumeric()
        ):
            # Numeric value is present in the config file
            seconds_expire = int(float(hours_expire) * 3600)
            if seconds_expire < longer_than:
                log.warning(
                    f"Invalid value for '{config_key}': {hours_expire}. Tokens"
                    f" must be valid for at least {longer_than} seconds. Using"
                    f" default value: {default_hours} hours"
                )
                if is_refresh:
                    log.warning(
                        "Note that refresh tokens should be valid at "
                        f"least {MIN_REFRESH_TOKEN_EXPIRY_DELTA} "
                        "seconds longer than access tokens."
                    )
                seconds_expire = int(float(default_hours) * 3600)
        else:
            # Non-numeric value is present in the config file. Warn and use
            # default
            log.error(
                f"Invalid value for '{config_key}': {hours_expire}. "
                f"Using default value: {default_hours} hours"
            )
            seconds_expire = int(float(default_hours) * 3600)

        return seconds_expire

    def configure_api(self) -> None:
        """Define global API output and its structure."""

        # helper to create HATEOAS schemas
        HATEOASModelSchema.api = self.api

        # whatever you get try to json it
        @self.api.representation("application/json")
        # pylint: disable=unused-argument
        def output_json(
            data: db.Base | list[db.Base], code: HTTPStatus, headers: dict = None
        ) -> Response:
            """
            Return jsonified data for request responses.

            Parameters
            ----------
            data: db.Base | list[db.Base]
                The data to be jsonified
            code: HTTPStatus
                The HTTP status code of the response
            headers: dict
                Additional headers to be added to the response
            """
            if isinstance(data, db.Base):
                data = jsonable(data)
            elif isinstance(data, list) and len(data) and isinstance(data[0], db.Base):
                data = jsonable(data)
            # Don't attempt json conversion if it's already a response.
            # Used for streaming endpoints where responses cannot be jsonified.
            elif isinstance(data, RequestsResponse):
                resp = make_response(data, code)
                return resp

            resp = make_response(json.dumps(data), code)
            resp.headers.extend(headers or {})
            return resp

    def configure_jwt(self):
        """Configure JWT authentication."""

        @self.jwt.additional_claims_loader
        # pylint: disable=unused-argument
        def additional_claims_loader(identity: db.Authenticatable | dict) -> dict:
            """
            Create additional claims for JWT tokens: set user type and for
            users, set their roles.

            Parameters
            ----------
            identity: db.Authenticatable | dict
                The identity for which to create the claims

            Returns
            -------
            dict:
                The claims to be added to the JWT token
            """
            roles = []
            if isinstance(identity, db.User):
                type_ = "user"
                roles = [role.name for role in identity.roles]

            elif isinstance(identity, db.Node):
                type_ = "node"
            elif isinstance(identity, dict):
                type_ = "container"
            else:
                log.error(f"could not create claims from {str(identity)}")
                return

            claims = {
                "client_type": type_,
                "roles": roles,
            }

            return claims

        @self.jwt.user_identity_loader
        # pylint: disable=unused-argument
        def user_identity_loader(identity: db.Authenticatable | dict) -> str | dict:
            """ "
            JSON serializing identity to be used by ``create_access_token``.

            Parameters
            ----------
            identity: db.Authenticatable | dict
                The identity to be serialized

            Returns
            -------
            str | dict:
                The serialized identity. For a node or user, this is the id;
                for a container, it is a dict.
            """
            if isinstance(identity, db.Authenticatable):
                return identity.id
            if isinstance(identity, dict):
                return identity

            log.error(
                f"Could not create a JSON serializable identity \
                        from '{str(identity)}'"
            )

        @self.jwt.user_lookup_loader
        # pylint: disable=unused-argument
        def user_lookup_loader(
            jwt_payload: dict, jwt_headers: dict
        ) -> db.Authenticatable | dict:
            """
            Load the user, node or container instance from the JWT payload.

            Parameters
            ----------
            jwt_payload: dict
                The JWT payload
            jwt_headers: dict
                The JWT headers

            Returns
            -------
            db.Authenticatable | dict:
                The user, node or container identity. If the identity is a
                container, a dict is returned.
            """
            identity = jwt_headers["sub"]
            auth_identity = Identity(identity)

            # in case of a user or node an auth id is shared as identity
            if isinstance(identity, int):
                # auth_identity = Identity(identity)

                auth = db.Authenticatable.get(identity)

                if isinstance(auth, db.Node):
                    for rule in db.Role.get_by_name(DefaultRole.NODE).rules:
                        auth_identity.provides.add(
                            RuleNeed(
                                name=rule.name,
                                scope=rule.scope,
                                operation=rule.operation,
                            )
                        )

                if isinstance(auth, db.User):
                    # add role permissions
                    for role in auth.roles:
                        for rule in role.rules:
                            auth_identity.provides.add(
                                RuleNeed(
                                    name=rule.name,
                                    scope=rule.scope,
                                    operation=rule.operation,
                                )
                            )

                    # add 'extra' permissions
                    for rule in auth.rules:
                        auth_identity.provides.add(
                            RuleNeed(
                                name=rule.name,
                                scope=rule.scope,
                                operation=rule.operation,
                            )
                        )

                identity_changed.send(
                    current_app._get_current_object(), identity=auth_identity
                )

                return auth
            else:
                # container identity

                for rule in db.Role.get_by_name(DefaultRole.CONTAINER).rules:
                    auth_identity.provides.add(
                        RuleNeed(
                            name=rule.name, scope=rule.scope, operation=rule.operation
                        )
                    )
                identity_changed.send(
                    current_app._get_current_object(), identity=auth_identity
                )
                log.debug("identity %s", identity)
                return identity

    def load_resources(self) -> None:
        """Import the modules containing API resources."""

        # make services available to the endpoints, this way each endpoint can
        # make use of 'em.
        services = {
            "socketio": self.socketio,
            "storage_adapter": self.storage_adapter,
            "mail": self.mail,
            "api": self.api,
            "permissions": self.permissions,
            "config": self.ctx.config,
        }

        for res in RESOURCES:
            module = importlib.import_module("vantage6.server.resource." + res)
            module.setup(self.api, self.ctx.config["api_path"], services)

    # TODO consider moving this method elsewhere. This is not trivial at the
    # moment because of the circular import issue with `db`, see
    # https://github.com/vantage6/vantage6/issues/53
    @staticmethod
    def _add_default_roles() -> None:
        for role in get_default_roles(db):
            # Normalize role name for logging and DB lookups: default-role
            # definitions use enum members, and this keeps the code robust if
            # plain strings are ever provided
            role_name = (
                role["name"].value
                if isinstance(role["name"], enum.Enum)
                else role["name"]
            )
            if not db.Role.get_by_name(role_name):
                log.warning("Creating new default role %s...", role_name)
                new_role = db.Role(
                    name=role_name,
                    description=role["description"],
                    rules=role["rules"],
                    is_default_role=True,
                )
                new_role.save()
            else:
                current_role = db.Role.get_by_name(role_name, is_default_role=True)
                if not current_role:
                    log.warning(
                        "'%s' is a default (built-in) role name. "
                        "A role by that name is already present in the database. "
                        "Will skip creating the default role by the same name.",
                        role_name,
                    )
                    continue
                # Check whether any default role properties changed. Use set()
                # to compare rules without relying on their order.
                has_rule_changes = set(current_role.rules) != set(role["rules"])
                has_description_change = current_role.description != role["description"]

                if has_rule_changes or has_description_change:
                    log.warning("Updating default role %s", role_name)
                    current_role.rules = role["rules"]
                    current_role.description = role["description"]
                    current_role.save()

    def start(self) -> None:
        """
        Start the server.

        Before server is really started, some database settings are checked and
        (re)set where appropriate.
        """
        # add default roles (if they don't exist yet)
        self._add_default_roles()

        # create root user if it is not in the DB yet
        try:
            db.User.get_by_username(SUPER_USER_INFO["username"])
        except Exception:
            log.warning("No root user found! Is this the first run?")
            self._create_super_user()

        return self

    def _create_super_user(self) -> None:
        """
        Create the super user.

        This method is used when the server is started for the first time.
        """
        # sanity check, this function should never be called in any other
        # context than the first run of the server
        try:
            db.User.get_by_username(SUPER_USER_INFO["username"])
            raise Exception("Attempted to create super user when it already existed!")
        except NoResultFound:
            pass

        log.debug("Creating organization for root user")
        org = db.Organization(name="root")

        # TODO use constant instead of 'Root' literal
        root = db.Role.get_by_name("Root")

        super_user_password = None

        # Note: server admin is supposed to set these env vars via docker
        # compose for instance
        if "V6_INIT_SUPER_PASS_HASHED_FILE" in os.environ:
            log.info("Using initial hashed super user password from file")

            with open(os.environ["V6_INIT_SUPER_PASS_HASHED_FILE"], "r") as f:
                super_user_password = HashedPassword(f.read().strip())

            if "V6_INIT_SUPER_PASS_HASHED" in os.environ:
                log.warn(
                    "Both V6_INIT_SUPER_PASS_HASHED_FILE and"
                    " V6_INIT_SUPER_PASS_HASHED are set. Using the value from"
                    " V6_INIT_SUPER_PASS_HASHED_FILE"
                )
        elif "V6_INIT_SUPER_PASS_HASHED" in os.environ:
            log.info(
                "Using initial hashed super user password provided via"
                " environemnt variable"
            )
            super_user_password = HashedPassword(
                os.environ["V6_INIT_SUPER_PASS_HASHED"]
            )
        else:
            # FIXME / TODO v5+: for backwards compatibility we still set this
            # as default but we might want to remove this soon!
            super_user_password = SUPER_USER_INFO["password"]
            log.warn(
                f"Creating super user ({SUPER_USER_INFO['username']})"
                " with default password!"
            )
            log.warn(
                "Please change it or set an initial password using the"
                " V6_INIT_SUPER_PASS_HASHED/_FILE environment variable"
            )

        # sanity check
        if not super_user_password:
            raise ValueError("No initial password assigned to root user!")

        user = db.User(
            username=SUPER_USER_INFO["username"],
            roles=[root],
            organization=org,
            # TODO: should we use RFC6761's "invalid." here?
            email="root@domain.ext",
            password=super_user_password,
            failed_login_attempts=0,
            last_login_attempt=None,
        )
        user.save()

    def __node_status_worker(self) -> None:
        """
        Set node status to offline if they haven't sent a ping message in a
        while.
        """
        # start periodic check if nodes are responsive
        while True:
            # Send ping event
            try:
                before_wait = dt.datetime.now(dt.timezone.utc)

                # Wait a while to give nodes opportunity to pong. This interval
                # is a bit longer than the interval at which the nodes ping,
                # because we want to make sure that the nodes have had time to
                # respond.
                time.sleep(PING_INTERVAL_SECONDS + 5)

                # Check for each node that is online if they have responded.
                # Otherwise set them to offline.
                online_status_nodes = db.Node.get_online_nodes()
                for node in online_status_nodes:
                    if node.last_seen.replace(tzinfo=dt.timezone.utc) < before_wait:
                        node.status = AuthStatus.OFFLINE.value
                        node.save()
            except Exception as e:
                log.exception("Node-status thread encountered an exception: %s", e)
                time.sleep(PING_INTERVAL_SECONDS)

    def __runs_data_cleanup_worker(self):
        """Start a background thread to clean up data from old Runs."""
        # NOTE/TODO: this is a very simple implementation, horizonal scaling is
        # not being taken into account. We'd probably only need one worker per
        # database, not per server instance (for example).
        include_input = self.ctx.config.get("runs_data_cleanup_include_input", False)
        while True:
            try:
                cleanup.cleanup_runs_data(
                    self.ctx.config,
                    include_input=include_input,
                )
            except Exception as e:
                log.error("Results cleanup failed. Will try again in one hour.")
                log.exception(e)
            # simple for now: check every hour
            time.sleep(3600)

    # TODO this functionality is temporarily disabled since it requires a user token
    # to couple the algorithm stores. It may be nice to find a way later to offer this
    # functionality again.
    # def couple_algorithm_stores(self) -> None:
    #     """Couple algorithm stores to the server.

    #     Checks if default algorithm stores are defined in the configuration and if so,
    #     couples them to the server.
    #     """
    #     algorithm_stores = self.ctx.config.get("algorithm_stores", [])
    #     server_url = get_server_url(self.ctx.config)
    #     if algorithm_stores and not server_url:
    #         log.warning(
    #             "Algorithm stores are defined in the configuration, but the server "
    #             "url is not. Skipping coupling of algorithm stores."
    #         )
    #         return
    #     if algorithm_stores:
    #         # TODO in the future it may change that not just any algorithm store can
    #         # be added to this server - in that case show a useful error below
    #         # (automated coupling is not possible for such cases I think?)

    #         # couple the stores
    #         for store in algorithm_stores:
    #             if not (name := store.get("name")):
    #                 log.warning("Algorithm store has no name, skipping coupling")
    #                 continue
    #             elif not (url := store.get("url")):
    #                 log.warning(
    #                     "Algorithm store %s has no url, skipping coupling", name
    #                 )
    #                 continue
    #             store = db.AlgorithmStore.get_by_url(url)
    #             if not store:
    #                 response, status = post_algorithm_store(
    #                     {
    #                         "name": name,
    #                         "algorithm_store_url": url,
    #                         "server_url": server_url,
    #                         "force": True,
    #                     },
    #                     self.ctx.config,
    #                 )
    #                 if status == HTTPStatus.CREATED:
    #                     log.info(
    #                         "Algorithm store '%s' at %s has been coupled to the server",
    #                         name,
    #                         url,
    #                     )
    #                 else:
    #                     log.error(
    #                         "Failed to couple algorithm store '%s' at %s to the server:"
    #                         " %s",
    #                         name,
    #                         url,
    #                         response["msg"],
    #                     )
    #             # else: store already exists, no need to couple it again


def run_server(config: str, system_folders: bool = True) -> ServerApp:
    """
    Run a vantage6 server.

    Parameters
    ----------
    config: str
        Configuration file path
    system_folders: bool
        Whether to use system or user folders. Default is True.

    Returns
    -------
    ServerApp
        A running instance of the vantage6 server
    """
    ctx = ServerContext.from_external_config_file(config, system_folders)
    allow_drop_all = ctx.config["allow_drop_all"]
    Database().connect(uri=ctx.get_database_uri(), allow_drop_all=allow_drop_all)
    return ServerApp(ctx).start()


def run_dev_server(server_app: ServerApp, *args, **kwargs) -> None:
    """
    Run a vantage6 development server (outside of a Docker container).

    Parameters
    ----------
    server_app: ServerApp
        Instance of a vantage6 server
    """
    log.warn("*" * 80)
    log.warn(" DEVELOPMENT SERVER ".center(80, "*"))
    log.warn("*" * 80)
    kwargs.setdefault("log_output", False)
    server_app.socketio.run(server_app.app, *args, **kwargs)



from http import HTTPStatus

import logging
from flask_restful import Api
from flask import request

from vantage6.common import logger_name
from vantage6.common.globals import BASIC_PROCESSING_IMAGE
from vantage6.server import db
from vantage6.server.permission import RuleCollection, Operation as P
from vantage6.server.resource import ServicesResources, with_user
from vantage6.server.resource.common.input_schema import ColumnNameInputSchema
from vantage6.server.resource.task import Tasks


module_name = logger_name(__name__)
log = logging.getLogger(module_name)


def setup(api: Api, api_base: str, services: dict) -> None:
    """
    Setup the column resource.

    Parameters
    ----------
    api : Api
        Flask restful api instance
    api_base : str
        Base url of the api
    services : dict
        Dictionary with services required for the resource endpoints
    """
    path = "/".join([api_base, module_name])
    log.info('Setting up "%s" and subdirectories', path)

    api.add_resource(
        ColumnNames,
        path,
        endpoint="column_names_without_id",
        methods=("POST",),
        resource_class_kwargs=services,
    )


# ------------------------------------------------------------------------------
# Resources / API's
# ------------------------------------------------------------------------------
column_name_input_schema = ColumnNameInputSchema()


class ColumnNames(ServicesResources):
    """Resource for requesting column names from a node database."""

    def __init__(self, socketio, storage_adapter, mail, api, permissions, config):
        super().__init__(socketio, storage_adapter, mail, api, permissions, config)
        # Use task permissions for this resource
        self.r_task: RuleCollection = getattr(self.permissions, "task")

    # TODO this endpoint currently requires the user to provide the
    # organization details (including input) for the task to retrieve the
    # column names. When the method name is stored in the database, and is thus
    # no longer encrypted, we can remove this requirement.
    @with_user
    def post(self):
        """Create request to get column names of a database. This is used in
        the user interface to show the user which columns are available.
        ---
        description: >-
          Returns a list of column names.\n

          Accessible to users based on their permissions to create tasks for
          the collaboration for which the request column names.

        requestBody:
          content:
            application/json:
              schema:
                properties:
                  db_label:
                    type: string
                    description: The label of the database whose column names
                      are requested
                  collaboration_id:
                    type: integer
                    description: The collaboration id for which database
                      columns are requested
                  organizations:
                    type: array
                    description: The organizations that should execute the
                      task. Each organization should be defined by a dictionary
                      with the `id` key and `input` data for a task.
                  sheet_name:
                    type: string
                    description: The name of the worksheet for which you want
                      to read the columns. Required for Excel files.
                  query:
                    type: string
                    description: The query that is used to obtain the data.
                      Required for SQL and SparQL databases.

        responses:
          201:
            description: Ok
          400:
            description: Non-allowed or wrong parameter values
          401:
            description: Unauthorized
          404:
            description: Collaboration not found

        security:
            - bearerAuth: []

        tags: ["Task"]
        """
        data = request.get_json()

        # validate request body
        errors = column_name_input_schema.validate(data)
        if errors:
            return {
                "msg": "Request body is incorrect",
                "errors": errors,
            }, HTTPStatus.BAD_REQUEST

        collaboration_id = data["collaboration_id"]
        collaboration = db.Collaboration.get(collaboration_id)
        if not collaboration:
            return {
                "msg": f"Collaboration id={collaboration_id} not found!"
            }, HTTPStatus.NOT_FOUND

        # check if user has permission to create a task for this collaboration:
        # that determines if the user is allowed to request column names
        if not self.r_task.can_for_col(P.CREATE, collaboration_id):
            return {
                "msg": "You lack the permission to do that!"
            }, HTTPStatus.UNAUTHORIZED

        # check if the column names are already available in the database
        # through shared node details.
        label = data["db_label"]
        colnames = []
        for node in collaboration.nodes:
            for record in node.config:
                if record.key == f"columns_{label}":
                    colnames.append(record.value)
            # note this indent: if one node has the column names, we assume
            # that all nodes have the column names. Otherwise it will just
            # lead to an error for the other nodes.
            if colnames:
                return {"columns": colnames}, HTTPStatus.OK

        # Column names not yet available, create new task to request column
        # names
        databases = [{"label": label}]
        if data.get("sheet_name"):
            databases[0]["sheet_name"] = data["sheet_name"]
        if data.get("query"):
            databases[0]["query"] = data["query"]
        return Tasks.post_task(
            data={
                "collaboration_id": collaboration_id,
                "name": "get_column_names",
                "image": BASIC_PROCESSING_IMAGE,
                "organizations": data["organizations"],
                "databases": databases,
            },
            socketio=self.socketio,
            rules=self.r_task,
            config=self.config,
        )



import logging

from flask import request, g
from flask_restful import Api
from http import HTTPStatus

from vantage6.common import logger_name
from vantage6.server import db
from vantage6.backend.common.resource.pagination import Pagination
from vantage6.server.permission import (
    Scope as S,
    Operation as P,
    PermissionManager,
    RuleCollection,
)
from vantage6.server.resource.common.input_schema import OrganizationInputSchema
from vantage6.server.resource import only_for, with_user, ServicesResources
from vantage6.server.resource.common.output_schema import OrganizationSchema


module_name = logger_name(__name__)
log = logging.getLogger(module_name)


def setup(api: Api, api_base: str, services: dict) -> None:
    """
    Setup the organization resource.

    Parameters
    ----------
    api : Api
        Flask restful api instance
    api_base : str
        Base url of the api
    services : dict
        Dictionary with services required for the resource endpoints
    """
    path = "/".join([api_base, module_name])
    log.info(f'Setting up "{path}" and subdirectories')

    api.add_resource(
        Organizations,
        path,
        endpoint="organization_without_id",
        methods=("GET", "POST"),
        resource_class_kwargs=services,
    )
    api.add_resource(
        Organization,
        path + "/<int:id>",
        endpoint="organization_with_id",
        methods=("GET", "PATCH", "DELETE"),
        resource_class_kwargs=services,
    )


# -----------------------------------------------------------------------------
# Permissions
# -----------------------------------------------------------------------------
def permissions(permissions: PermissionManager) -> None:
    """
    Define the permissions for this resource.

    Parameters
    ----------
    permissions : PermissionManager
        Permission manager instance to which permissions are added
    """
    add = permissions.appender(module_name)

    add(scope=S.GLOBAL, operation=P.VIEW, description="view any organization")
    add(
        scope=S.ORGANIZATION,
        operation=P.VIEW,
        description="view your own organization info",
        assign_to_container=True,
        assign_to_node=True,
    )
    add(
        scope=S.COLLABORATION,
        operation=P.VIEW,
        description="view collaborating organizations",
        assign_to_container=True,
        assign_to_node=True,
    )
    add(scope=S.GLOBAL, operation=P.EDIT, description="edit any organization")
    add(
        scope=S.ORGANIZATION,
        operation=P.EDIT,
        description="edit your own organization info",
        assign_to_node=True,
    )
    add(
        scope=S.COLLABORATION,
        operation=P.EDIT,
        description="edit collaborating organizations",
    )
    add(scope=S.GLOBAL, operation=P.CREATE, description="create a new organization")
    add(scope=S.GLOBAL, operation=P.DELETE, description="delete any organization")


# ------------------------------------------------------------------------------
# Resources / API's
# ------------------------------------------------------------------------------
org_schema = OrganizationSchema()
org_input_schema = OrganizationInputSchema()


class OrganizationBase(ServicesResources):
    def __init__(self, socketio, storage_adapter, mail, api, permissions, config):
        super().__init__(socketio, storage_adapter, mail, api, permissions, config)
        self.r: RuleCollection = getattr(self.permissions, module_name)


class Organizations(OrganizationBase):
    @only_for(("user", "node", "container"))
    def get(self):
        """Returns a list organizations
        ---
        description: >-
            Get a list of organizations based on filters and user permissions\n

            ### Permission Table\n
            |Rule name|Scope|Operation|Assigned to node|Assigned to container|
            Description|\n
            |--|--|--|--|--|--|\n
            |Organization|Global|View|❌|❌|View all organizations|\n
            |Organization|Collaboration|View|✅|✅|View a list of organizations
            within the scope of the collaboration|\n
            |Organization|Organization|View|✅|✅|View a 'list' of just the
            organization you are part of|\n

            Accessible to users.

        parameters:
          - in: query
            name: name
            schema:
              type: string
            description: >-
              Name to match with a LIKE operator. \n
              * The percent sign (%) represents zero, one, or multiple
              characters\n
              * underscore sign (_) represents one, single character
          - in: query
            name: country
            schema:
              type: string
            description: Country
          - in: query
            name: collaboration_id
            schema:
              type: integer
            description: Collaboration id
          - in: query
            name: study_id
            schema:
              type: integer
            description: Study id
          - in: query
            name: ids
            schema:
              type: array
              items:
                type: integer
            description: List of organization ids. Only these organizations will be
              returned
          - in: query
            name: page
            schema:
              type: integer
            description: Page number for pagination (default=1)
          - in: query
            name: per_page
            schema:
              type: integer
            description: Number of items per page (default=10)
          - in: query
            name: sort
            schema:
              type: string
            description: >-
              Sort by one or more fields, separated by a comma. Use a minus
              sign (-) in front of the field to sort in descending order.

        responses:
          200:
            description: Ok
          401:
            description: Unauthorized
          400:
            description: Improper values for pagination or sorting parameters

        security:
          - bearerAuth: []

        tags: ["Organization"]
        """

        # Obtain the organization of the requester
        auth_org = self.obtain_auth_organization()
        args = request.args

        # query
        q = g.session.query(db.Organization)
        g.session.commit()

        # filter by a field of this endpoint
        if "name" in args:
            q = q.filter(db.Organization.name.like(args["name"]))
        if "country" in args:
            q = q.filter(db.Organization.country == args["country"])
        if "collaboration_id" in args:
            if not self.r.can_for_col(P.VIEW, int(args["collaboration_id"])):
                return {
                    "msg": "You lack the permission to get all organizations "
                    "in your collaboration!"
                }, HTTPStatus.UNAUTHORIZED
            q = (
                q.join(db.Member)
                .join(db.Collaboration)
                .filter(db.Collaboration.id == args["collaboration_id"])
            )
        if "study_id" in args:
            study = db.Study().get(args["study_id"])
            if not self.r.can_for_col(P.VIEW, study.collaboration_id):
                return {
                    "msg": "You lack the permission to get all organizations "
                    "in a study!"
                }, HTTPStatus.UNAUTHORIZED
            q = (
                q.join(db.StudyMember)
                .join(db.Study)
                .filter(db.Study.id == args["study_id"])
            )
        # filter by a list of organization ids. This option is mostly used in the UI,
        # where it is convenient to get details on a list of organizations at once.
        if "ids" in args:
            q = q.filter(db.Organization.id.in_(args.getlist("ids")))

        # filter the list of organizations based on the scope
        if self.r.v_glo.can():
            pass  # don't apply filters
        elif self.r.v_col.can():
            # obtain collaborations your organization participates in
            collabs = (
                g.session.query(db.Collaboration)
                .filter(db.Collaboration.organizations.any(id=auth_org.id))
                .all()
            )
            g.session.commit()

            # filter orgs in own collaborations, and add own organization in
            # case this organization does not participate in any collaborations
            # yet
            org_ids = [o.id for col in collabs for o in col.organizations]
            org_ids = list(set(org_ids + [auth_org.id]))

            # select only the organizations in the collaborations
            q = q.filter(db.Organization.id.in_(org_ids))

        elif self.r.v_org.can():
            q = q.filter(db.Organization.id == auth_org.id)
        else:
            return {
                "msg": "You lack the permission to do that!"
            }, HTTPStatus.UNAUTHORIZED

        # paginate the results
        try:
            page = Pagination.from_query(q, request, db.Organization)
        except (ValueError, AttributeError) as e:
            return {"msg": str(e)}, HTTPStatus.INTERNAL_SERVER_ERROR

        # serialization of DB model
        return self.response(page, org_schema)

    @with_user
    def post(self):
        """Create new organization
        ---
        description: >-
          Creates a new organization from the specified values\n

          ### Permission Table\n
          |Rule name|Scope|Operation|Assigned to node|Assigned to container|
          Description|\n
          |--|--|--|--|--|--|\n
          |Organization|Global|Create|❌|❌|Create a new organization|\n

          Accessible to users.

        requestBody:
          content:
            application/json:
              schema:
                $ref: '#/components/schemas/Organization'

        responses:
          201:
            description: Ok
          401:
            description: Unauthorized
          400:
            description: Organization with that name already exists

        security:
          - bearerAuth: []

        tags: ["Organization"]
        """

        if not self.r.c_glo.can():
            return {
                "msg": "You lack the permissions to do that!"
            }, HTTPStatus.UNAUTHORIZED

        # validate request body
        data = request.get_json()
        errors = org_input_schema.validate(data)
        if errors:
            return {
                "msg": "Request body is incorrect",
                "errors": errors,
            }, HTTPStatus.BAD_REQUEST

        name = data.get("name")
        if db.Organization.exists("name", name):
            return {
                "msg": f"Organization with name '{name}' already exists!"
            }, HTTPStatus.BAD_REQUEST

        organization = db.Organization(
            name=name,
            address1=data.get("address1", ""),
            address2=data.get("address2" ""),
            zipcode=data.get("zipcode", ""),
            country=data.get("country", ""),
            public_key=data.get("public_key", ""),
            domain=data.get("domain", ""),
        )
        organization.save()

        return org_schema.dump(organization, many=False), HTTPStatus.CREATED


class Organization(OrganizationBase):
    @only_for(("user", "node", "container"))
    def get(self, id):
        """Get organization
        ---
        description: >-
          Returns the organization specified by the id\n

          ### Permission Table\n
          |Rule name|Scope|Operation|Assigned to node|Assigned to container|
          Description|\n
          |--|--|--|--|--|--|\n
          |Organization|Global|View|❌|❌|View all organizations|\n
          |Organization|Collaboration|View|✅|✅|View a list of organizations
          within the scope of the collaboration|\n
          |Organization|Organization|View|✅|✅|View a list of organizations
          that the user is part of|\n

          Accessible to users.

        parameters:
          - in: path
            name: id
            schema:
              type: integer
            description: Organization id
            required: true

        responses:
          200:
            description: Ok
          404:
            description: Organization not found
          401:
            description: Unauthorized

        security:
          - bearerAuth: []

        tags: ["Organization"]
        """

        # retrieve requested organization
        req_org = db.Organization.get(id)
        if not req_org:
            return {"msg": f"Organization id={id} not found!"}, HTTPStatus.NOT_FOUND

        # Check if auth has enough permissions
        if not self.r.allowed_for_org(P.VIEW, id):
            return {
                "msg": "You do not have permission to do that!"
            }, HTTPStatus.UNAUTHORIZED

        return org_schema.dump(req_org, many=False), HTTPStatus.OK

    @only_for(("user", "node"))
    def patch(self, id):
        """Update organization
        ---
        description: >-
          Updates the organization with the specified id.\n

          ### Permission Table\n
          |Rule name|Scope|Operation|Assigned to node|Assigned to container|
          Description|\n
          |--|--|--|--|--|--|\n
          |Organization|Global|Edit|❌|❌|Update an organization with
          specified id|\n
          |Organization|Collaboration|Edit|❌|❌|Update an organization within
          the collaboration the user is part of|\n
          |Organization|Organization|Edit|❌|❌|Update the organization that
          the user is part of|\n

          Accessible to users.

        parameters:
          - in: path
            name: id
            schema:
              type: integer
            description: Organization id
            required: tr

        requestBody:
          content:
            application/json:
              schema:
                $ref: '#/components/schemas/Organization'

        responses:
          200:
            description: Ok
          404:
            description: Organization with specified id is not found
          401:
            description: Unauthorized
          400:
            description: Organization with that name already exists

        security:
          - bearerAuth: []

        tags: ["Organization"]
        """
        # validate request body
        data = request.get_json()
        errors = org_input_schema.validate(data, partial=True)
        if errors:
            return {
                "msg": "Request body is incorrect",
                "errors": errors,
            }, HTTPStatus.BAD_REQUEST

        organization = db.Organization.get(id)
        if not organization:
            return {"msg": f"Organization with id={id} not found"}, HTTPStatus.NOT_FOUND

        if not self.r.allowed_for_org(P.EDIT, id):
            return {
                "msg": "You lack the permission to do that!"
            }, HTTPStatus.UNAUTHORIZED

        name = data.get("name")
        if name:
            if organization.name != name and db.Organization.exists("name", name):
                return {
                    "msg": f"Organization with name '{name}' already exists!"
                }, HTTPStatus.BAD_REQUEST
            organization.name = name

        fields = ["address1", "address2", "zipcode", "country", "public_key", "domain"]
        for field in fields:
            if field in data and data[field] is not None:
                setattr(organization, field, data[field])

        organization.save()
        return org_schema.dump(organization, many=False), HTTPStatus.OK

    @with_user
    def delete(self, id):
        """Delete organization
        ---
        description: >-
          Deletes the organization with the specified id\n

          ### Permission Table\n
          |Rule name|Scope|Operation|Assigned to node|Assigned to container|
          Description|\n
          |--|--|--|--|--|--|\n
          |Organization|Global|Delete|❌|❌|Delete an organization with
          specified id|\n

          Accessible to users.

        parameters:
          - in: path
            name: id
            schema:
              type: integer
            description: Organization id
            required: true
          - in: query
            name: delete_dependents
            schema:
              type: boolean
            description: >-
              If true, also delete all dependents (i.e. users, roles, tasks, runs,
              nodes). Collaborations and studies are not deleted, but the organization
              is removed from them.

        responses:
          200:
            description: Ok
          404:
            description: Organization with specified id is not found
          401:
            description: Unauthorized
          400:
            description: Organization has dependents or user tries to delete own
              organization


        security:
          - bearerAuth: []

        tags: ["Organization"]
        """
        if not self.r.d_glo.can():
            return {
                "msg": "You lack the permission to do that!"
            }, HTTPStatus.UNAUTHORIZED

        organization = db.Organization.get(id)
        if not organization:
            return {"msg": f"Organization with id={id} not found"}, HTTPStatus.NOT_FOUND

        if g.user.organization_id == id:
            return {
                "msg": "You cannot delete your own organization!"
            }, HTTPStatus.BAD_REQUEST

        if (
            organization.runs
            or organization.tasks
            or organization.nodes
            or organization.users
            or organization.roles
        ):
            delete_dependents = request.args.get("delete_dependents", False)
            if not delete_dependents:
                return {
                    "msg": (
                        f"Organization has dependents: {len(organization.runs)} runs, "
                        f"{len(organization.tasks)} tasks, {len(organization.nodes)} "
                        f"nodes, {len(organization.users)} users, and "
                        f"{len(organization.roles)} roles. Please specify "
                        "'delete_dependents=true' to delete them."
                    )
                }, HTTPStatus.BAD_REQUEST

            # delete dependents
            if organization.runs:
                log.warning(
                    "Deleting %s runs of organization %s",
                    len(organization.runs),
                    organization.name,
                )
                for run in organization.runs:
                    run.delete()
            if organization.tasks:
                log.warning(
                    "Deleting %s tasks of organization %s",
                    len(organization.tasks),
                    organization.name,
                )
                for task in organization.tasks:
                    task.delete()
            if organization.nodes:
                log.warning(
                    "Deleting %s nodes of organization %s",
                    len(organization.nodes),
                    organization.name,
                )
                for node in organization.nodes:
                    node.delete()
            if organization.users:
                log.warning(
                    "Deleting %s users of organization %s",
                    len(organization.users),
                    organization.name,
                )
                for user in organization.users:
                    user.delete()
            if organization.roles:
                log.warning(
                    "Deleting %s roles of organization %s",
                    len(organization.roles),
                    organization.name,
                )
                for role in organization.roles:
                    role.delete()

        # remove organization from collaborations and studies
        for study in organization.studies:
            study.organizations.remove(organization)
        for collaboration in organization.collaborations:
            collaboration.organizations.remove(organization)
        organization.delete()

        return {
            "msg": f"Organization id={id} was removed from the database"
        }, HTTPStatus.OK



import bcrypt

from sqlalchemy import Column, String, DateTime

from vantage6.server.model.base import Base


class Authenticatable(Base):
    """
    Parent table of database entities that can authenticate.

    Entities that can authenticate are nodes and users. Containers
    can also authenticate but these are authenticated indirectly
    through the nodes.
    """

    # fields
    type = Column(String(50))
    __mapper_args__ = {
        "polymorphic_identity": "authenticatable",
        "polymorphic_on": type,
    }
    ip = Column(String)
    last_seen = Column(DateTime)
    status = Column(String)

    @staticmethod
    def hash(secret: str) -> str:
        """
        Hash a secret using bcrypt.

        Parameters
        ----------
        secret : str
            Secret to be hashed

        Returns
        -------
        str
            Hashed secret
        """
        return bcrypt.hashpw(secret.encode("utf8"), bcrypt.gensalt()).decode("utf8")



from __future__ import annotations
import bcrypt

from sqlalchemy.orm import relationship, validates
from sqlalchemy import Column, Integer, String, ForeignKey

from vantage6.common.globals import AuthStatus
from vantage6.server.model.base import DatabaseSessionManager
from vantage6.server.model.authenticatable import Authenticatable


class Node(Authenticatable):
    """
    Table that contains all registered nodes.

    Attributes
    ----------
    id : int
        Primary key
    name : str
        Name of the node
    api_key : str
        API key of the node
    collaboration : :class:`~.model.collaboration.Collaboration`
        Collaboration that the node belongs to
    organization : :class:`~.model.organization.Organization`
        Organization that the node belongs to
    """

    _hidden_attributes = ["api_key"]

    id = Column(Integer, ForeignKey("authenticatable.id"), primary_key=True)

    # fields
    name = Column(String, unique=True)
    api_key = Column(String)
    collaboration_id = Column(Integer, ForeignKey("collaboration.id"))
    organization_id = Column(Integer, ForeignKey("organization.id"))

    # relationships
    collaboration = relationship("Collaboration", back_populates="nodes")
    organization = relationship("Organization", back_populates="nodes")
    config = relationship("NodeConfig", back_populates="node")

    # the type specification in Authenticatable
    __mapper_args__ = {
        "polymorphic_identity": "node",
    }

    @validates("api_key")
    def _validate_api_key(self, key: str, api_key: str) -> str:
        """
        Hashes the api_key before storing it in the database.

        Parameters
        ----------
        key : str
            The name of the attribute that is being validated
        api_key : str
            The value of the attribute that is being validated

        Returns
        -------
        str
            The hashed api_key
        """
        return self.hash(api_key)

    def check_key(self, key: str) -> bool:
        """
        Checks if the provided key matches the stored key.

        Parameters
        ----------
        key : str
            The key to check

        Returns
        -------
        bool
            True if the provided key matches the stored key, False otherwise
        """
        if self.api_key is not None:
            expected_hash = self.api_key.encode("utf8")
            return bcrypt.checkpw(key.encode("utf8"), expected_hash)
        return False

    @classmethod
    def get_by_api_key(cls, api_key: str) -> Node | None:
        """
        Returns Node based on the provided API key.

        Parameters
        ----------
        api_key : str
            The API key of the node to search for

        Returns
        -------
        Node | None
            Returns the node if a node is associated with api_key, None if no
            node is associated with api_key.
        """
        session = DatabaseSessionManager.get_session()

        nodes = session.query(cls).all()
        session.commit()
        for node in nodes:
            is_correct_key = node.check_key(api_key)
            if is_correct_key:
                return node
        # no node found with matching API key
        return None

    @classmethod
    def get_online_nodes(cls) -> list[Node]:
        """
        Return nodes that currently have status 'online'

        Returns
        -------
        list[Node]
            List of node models that are currently online
        """
        session = DatabaseSessionManager.get_session()

        result = session.query(cls).filter_by(status=AuthStatus.ONLINE.value).all()
        session.commit()
        return result

    @classmethod
    def exists_by_id(cls, organization_id: int, collaboration_id: int) -> bool:
        """
        Check if a node exists for the given organization and collaboration.

        Parameters
        ----------
        organization_id : int
            The id of the organization
        collaboration_id : int
            The id of the collaboration

        Returns
        -------
        bool
            True if a node exists for the given organization and collaboration,
            False otherwise.
        """
        session = DatabaseSessionManager.get_session()
        result = (
            session.query(cls)
            .filter_by(
                organization_id=organization_id, collaboration_id=collaboration_id
            )
            .scalar()
        )
        session.commit()
        return result

    def __repr__(self) -> str:
        """
        String representation of the Node model.

        Returns
        -------
        str
            String representation of the Node model
        """
        return (
            "<Node "
            f"{self.id}: '{self.name}', "
            f"organization: {self.organization.name}, "
            f"collaboration: {self.collaboration.name}, "
            f"api_key: {self.api_key} "
            ">"
        )



"""
The role_rule_assocation table defines which rules have been assigned to which
roles. Each line contains a rule_id that is a member of a certain role_id. Each
role will usually have multiple rules assigned to it.
"""

from sqlalchemy import Column, Integer, ForeignKey, Table

from .base import Base

# TODO: refactor to use the server model
role_rule_association = Table(
    "role_rule_association",
    Base.metadata,
    Column("role_id", Integer, ForeignKey("role.id")),
    Column("rule_id", Integer, ForeignKey("rule.id")),
)



import unittest

from datetime import datetime, timedelta, timezone
import uuid
from unittest.mock import patch, call

from vantage6.server.model.run import Run
from vantage6.server.model.base import Database, DatabaseSessionManager
from vantage6.server.controller import cleanup
from vantage6.common.task_status import TaskStatus
from vantage6.server.model import Task


class TestCleanupRunsIsolated(unittest.TestCase):
    def setUp(self):
        Database().connect("sqlite://", allow_drop_all=True)
        self.session = DatabaseSessionManager.get_session()

        self.uuid = str(uuid.uuid4())

    def tearDown(self):
        Database().clear_data()

    @patch("vantage6.server.model.task.Task")
    def test_cleanup_completed_old_run(self, mock_task):
        # Now use the actual instances, not ints
        task = Task(
            name="test-task",
            description="Test task for cleanup",
            image="test-image:latest",
        )
        self.session.add(task)
        self.session.commit()

        run = Run(
            finished_at=datetime.now(timezone.utc) - timedelta(days=31),
            result=self.uuid,
            input="input",
            log="log should be preserved",
            status=TaskStatus.COMPLETED,
            task=task,
        )

        self.session.add(run)
        self.session.commit()

        cleanup.cleanup_runs_data({"runs_data_cleanup_days": 30}, include_input=True)
        self.session.refresh(run)
        self.assertEqual(run.result, "")
        self.assertEqual(run.input, "")
        self.assertEqual(run.log, "log should be preserved")
        self.assertIsNotNone(run.cleanup_at)

    @patch(
        "vantage6.server.service.azure_storage_service.AzureStorageService.delete_blob"
    )
    def test_cleanup_completed_old_blob(self, mock_delete_blob):
        task = Task(
            name="test-task",
            description="Test task for cleanup",
            image="test-image:latest",
        )
        self.session.add(task)
        self.session.commit()

        run = Run(
            finished_at=datetime.now(timezone.utc) - timedelta(days=31),
            result=self.uuid,
            input="input",
            log="log should be preserved",
            status=TaskStatus.COMPLETED,
            task=task,
            blob_storage_used=True,
        )

        config = {
            "runs_data_cleanup_days": 30,
            "large_result_store": {
                "type": "azure",
                "container_name": "test-container",
                "connection_string": "DefaultEndpointsProtocol=https;AccountName=dummyname;AccountKey=dummykey",
            },
        }

        self.session.add(run)
        self.session.commit()

        cleanup.cleanup_runs_data(config, include_input=True)
        self.session.refresh(run)

        expected_calls = [call(self.uuid), call("input")]
        mock_delete_blob.assert_has_calls(expected_calls, any_order=False)

    def test_no_cleanup_recent_completed_run(self):
        # Ineligible: completed, but not old enough

        run = Run(
            finished_at=datetime.now(timezone.utc) - timedelta(days=10),
            result=self.uuid,
            input="input",
            status=TaskStatus.COMPLETED,
        )
        self.session.add(run)
        self.session.commit()

        cleanup.cleanup_runs_data({"runs_data_cleanup_days": 30}, include_input=True)
        self.session.refresh(run)
        self.assertEqual(run.result, self.uuid)
        self.assertEqual(run.input, "input")
        self.assertIsNone(run.cleanup_at)

    def test_no_cleanup_non_completed_run(self):
        # Ineligible: not COMPLETED, albeit old enough
        run = Run(
            finished_at=datetime.now(timezone.utc) - timedelta(days=31),
            result=self.uuid,
            input="input",
            status=TaskStatus.FAILED,  # Not COMPLETED, so ineligible
        )
        self.session.add(run)
        self.session.commit()

        cleanup.cleanup_runs_data({"runs_data_cleanup_days": 30}, include_input=True)
        self.session.refresh(run)
        self.assertEqual(run.result, self.uuid)
        self.assertEqual(run.input, "input")
        self.assertIsNone(run.cleanup_at)

    def test_cleanup_without_clearing_input(self):
        # Eligible: completed > 30 days ago
        run = Run(
            finished_at=datetime.now(timezone.utc) - timedelta(days=40),
            result=self.uuid,
            input="input",
            status=TaskStatus.COMPLETED,
        )
        self.session.add(run)
        self.session.commit()

        cleanup.cleanup_runs_data({"runs_data_cleanup_days": 30}, include_input=False)
        self.session.refresh(run)
        self.assertEqual(run.result, "")
        self.assertEqual(run.input, "input")
        self.assertIsNotNone(run.cleanup_at)


# Test cleanup_results()' handling of multiple runs
class TestCleanupRunsCount(unittest.TestCase):
    def setUp(self):
        Database().connect("sqlite://", allow_drop_all=True)
        self.session = DatabaseSessionManager.get_session()

    def tearDown(self):
        # clear_data() will clear session too
        Database().clear_data()

    def create_runs(self):
        run0 = Run(
            finished_at=datetime.now(timezone.utc) - timedelta(days=31),
            result="result0",
            input="input0",
            status=TaskStatus.COMPLETED,
        )
        run1 = Run(
            finished_at=datetime.now(timezone.utc) - timedelta(days=200),
            result="result1",
            input="input1",
            status=TaskStatus.COMPLETED,
        )
        run2 = Run(
            finished_at=datetime.now(timezone.utc) - timedelta(days=10),
            result="result2",
            input="input2",
            status=TaskStatus.COMPLETED,
        )
        run3 = Run(
            finished_at=datetime.now(timezone.utc) - timedelta(days=10),
            result="result3",
            input="input3",
            status=TaskStatus.PENDING,
        )
        run4 = Run(
            finished_at=datetime.now(timezone.utc) - timedelta(days=31),
            result="result4",
            input="input4",
            status=TaskStatus.FAILED,
        )
        run5 = Run(
            finished_at=datetime.now(timezone.utc) - timedelta(days=10),
            result="result5",
            input="input5",
            status=TaskStatus.ACTIVE,
        )
        self.session.add_all([run0, run1, run2, run3, run4, run5])
        self.session.commit()

        return run0, run1, run2, run3, run4, run5

    def test_cleanup_runs_count(self):

        # Insert runs into db
        runs = self.create_runs()

        # clean up result and input from runs older than 30 days
        cleanup.cleanup_runs_data({"runs_data_cleanup_days": 30}, include_input=True)

        # db has been changed, refresh objects
        for run in runs:
            self.session.refresh(run)

        # Query the DB to get cleaned up and non-cleaned up runs
        cleaned_runs = self.session.query(Run).filter(Run.cleanup_at != None).all()
        remaining_runs = self.session.query(Run).filter(Run.cleanup_at == None).all()

        # We expect only run0 and run1 to have been cleaned up
        self.assertEqual(len(cleaned_runs), 2)
        self.assertEqual(len(remaining_runs), 4)

        # Verify that cleaned runs have their result and input cleared, and
        # cleanup_at set
        for run in cleaned_runs:
            self.assertEqual(run.result, "")
            self.assertEqual(run.input, "")
            self.assertIsNotNone(run.cleanup_at)

        # Verify that runs that should not have been cleaned up keep their
        # result and input
        for i in range(2, 6):
            self.assertEqual(runs[i].result, f"result{i}")
            self.assertEqual(runs[i].input, f"input{i}")



import logging
import datetime as dt
from http import HTTPStatus

import gevent
from flask import request, render_template, g, current_app, Flask
from flask_jwt_extended import create_access_token, decode_token
from flask_restful import Api
from flask_mail import Mail
from jwt.exceptions import DecodeError
from sqlalchemy.orm.exc import NoResultFound

from vantage6.common import logger_name, generate_apikey
from vantage6.common.globals import APPNAME, MAIN_VERSION_NAME
from vantage6.backend.common.globals import (
    DEFAULT_EMAIL_FROM_ADDRESS,
    DEFAULT_SUPPORT_EMAIL_ADDRESS,
)
from vantage6.server import db
from vantage6.server.globals import (
    DEFAULT_EMAILED_TOKEN_VALIDITY_MINUTES,
    DEFAULT_BETWEEN_USER_EMAILS_MINUTES,
)
from vantage6.server.model.rule import Operation
from vantage6.server.resource import ServicesResources, with_user
from vantage6.server.resource.common.auth_helper import create_qr_uri, user_login
from vantage6.server.resource.common.input_schema import (
    ChangePasswordInputSchema,
    RecoverPasswordInputSchema,
    ResetPasswordInputSchema,
    Recover2FAInputSchema,
    Reset2FAInputSchema,
    ResetAPIKeyInputSchema,
)
from vantage6.server.model.user import User

module_name = logger_name(__name__)
log = logging.getLogger(module_name)


def setup(api: Api, api_base: str, services: dict) -> None:
    """
    Setup the recover resource.

    Parameters
    ----------
    api : Api
        Flask restful api instance
    api_base : str
        Base url of the api
    services : dict
        Dictionary with services required for the resource endpoints
    """
    path = "/".join([api_base, module_name])
    log.info(f'Setting up "{path}" and subdirectories')

    api.add_resource(
        ResetPassword,
        path + "/reset",
        endpoint="reset_password",
        methods=("POST",),
        resource_class_kwargs=services,
    )

    api.add_resource(
        RecoverPassword,
        path + "/lost",
        endpoint="recover_password",
        methods=("POST",),
        resource_class_kwargs=services,
    )

    api.add_resource(
        ResetTwoFactorSecret,
        path + "/2fa/reset",
        endpoint="reset_2fa_secret",
        methods=("POST",),
        resource_class_kwargs=services,
    )

    api.add_resource(
        RecoverTwoFactorSecret,
        path + "/2fa/lost",
        endpoint="recover_2fa_secret",
        methods=("POST",),
        resource_class_kwargs=services,
    )

    api.add_resource(
        ChangePassword,
        api_base + "/password/change",
        endpoint="change_password",
        methods=("PATCH",),
        resource_class_kwargs=services,
    )

    api.add_resource(
        ResetAPIKey,
        path + "/node",
        endpoint="reset_api_key",
        methods=("POST",),
        resource_class_kwargs=services,
    )


recover_pw_schema = RecoverPasswordInputSchema()
reset_pw_schema = ResetPasswordInputSchema()
recover_2fa_schema = Recover2FAInputSchema()
reset_2fa_schema = Reset2FAInputSchema()
reset_api_key_schema = ResetAPIKeyInputSchema()
change_pw_schema = ChangePasswordInputSchema()


# used by RecoverPassword.post()
def _handle_password_recovery(
    app: Flask, username: str, email: str, config: dict, mail: Mail
) -> None:
    """
    Send an email to user with a password reset token.

    This function also checks whether such an email has been sent recently, and
    if so avoids sending it.

    Parameters
    ----------
    app: flask.Flask
        The current Flask app
    username: str
        User for who the password reset is being requested
    email: str
        Email address associated to an account for which the password reset is
        being requested
    config: dict
        Dictionary with configuration settings
    mail: flask_mail.Mail
        An instance of the Flask mail class. Used to send email to user in case
        of too many failed login attempts.
    """
    # read settings
    password_policy = config.get("password_policy", {})
    minutes_between_password_reset_emails = password_policy.get(
        "between_user_emails_minutes",
        DEFAULT_BETWEEN_USER_EMAILS_MINUTES,
    )
    smtp_settings = config.get("smtp")
    if not smtp_settings:
        log.warning("No SMTP settings found in config - cannot send email!")
        return
    minutes_token_valid = smtp_settings.get(
        "email_token_validity_minutes", DEFAULT_EMAILED_TOKEN_VALIDITY_MINUTES
    )
    expires = dt.timedelta(minutes=minutes_token_valid)
    email_from = smtp_settings.get("email_from", DEFAULT_EMAIL_FROM_ADDRESS)
    support_email = config.get("support_email", DEFAULT_SUPPORT_EMAIL_ADDRESS)

    try:
        user = User.get_by_username(username) if username else User.get_by_email(email)
    except NoResultFound:
        account_name = username or email
        log.info(
            "Someone requested password recovery for non-existing account '%s'",
            account_name,
        )
        return

    log.info("Password reset requested for '%s'", user.username)

    # check that email has not already been sent recently
    email_sent_recently = user.last_email_recover_password_sent and (
        dt.datetime.now(dt.timezone.utc)
        < user.last_email_recover_password_sent
        + dt.timedelta(minutes=minutes_between_password_reset_emails)
    )
    if email_sent_recently:
        log.info("Skipping sending password reset email to '%s'", user.username)
        return

    with app.app_context():
        # generate a token that can reset their password
        reset_token = create_access_token({"id": str(user.id)}, expires_delta=expires)
        log.info("Sending password reset email to '%s'", user.email)
        mail.send_email(
            f"Password reset {APPNAME}",
            sender=email_from,
            recipients=[user.email],
            text_body=render_template(
                "mail/reset_token.txt",
                token=reset_token,
                firstname=user.firstname,
                reset_type="password",
                what_to_do="simply ignore this message",
            ),
            html_body=render_template(
                "mail/reset_token.html",
                token=reset_token,
                firstname=user.firstname,
                reset_type="password",
                support_email=support_email,
                what_to_do="simply ignore this message",
            ),
        )

    # Update last password reset email sent date
    user.last_email_recover_password_sent = dt.datetime.now(dt.timezone.utc)
    user.save()


# ------------------------------------------------------------------------------
# Resources / API's
# ------------------------------------------------------------------------------
class ResetPassword(ServicesResources):
    """User can use recover token to reset their password."""

    def post(self):
        """Set a new password using a recover token
        ---
        description: >-
          User can use a recover token to reset their password

        requestBody:
          content:
            application/json:
              schema:
                properties:
                  reset_token:
                    type: string
                    description: Recover token (received by email)
                  password:
                    type: string
                    description: New password

        responses:
          200:
            description: Ok
          400:
            description: Password or recovery token is missing or invalid

        tags: ["Account recovery"]
        """
        body = request.get_json()
        # validate request body
        errors = reset_pw_schema.validate(body)
        if errors:
            return {
                "msg": "Request body is incorrect",
                "errors": errors,
            }, HTTPStatus.BAD_REQUEST

        reset_token = body.get("reset_token")
        password = body.get("password")

        # obtain user
        try:
            user_id = decode_token(reset_token)["sub"].get("id")
        except DecodeError:
            return {"msg": "Invalid recovery token!"}, HTTPStatus.BAD_REQUEST

        user = db.User.get(user_id)

        # reset number of failed login attempts to prevent that user cannot
        # reactivate via email
        user.failed_login_attempts = 0
        user.save()

        # set password
        msg = user.set_password(password)
        if msg:
            return {"msg": msg}, HTTPStatus.BAD_REQUEST

        log.info("Successful password reset for '%s'", user.username)
        return {"msg": "The password has successfully been reset!"}, HTTPStatus.OK


class RecoverPassword(ServicesResources):
    """Send a mail containing a recover token to reset the password"""

    def post(self):
        """Request a recover token
        ---
        description: >-
          Request a recover token if password is lost. Either email address
          or username must be supplied.

        requestBody:
          content:
            application/json:
              schema:
                properties:
                  username:
                    type: string
                    description: Username from which the password needs to be
                      recovered
                  email:
                    type: string
                    description: Email of user from which the password needs to
                      be recovered

        responses:
          200:
            description: Ok
          400:
            description: No username or email provided

        tags: ["Account recovery"]
        """
        # default return string
        ret = {
            "msg": "If the username or email is in our database you "
            "will soon receive an email."
        }
        body = request.get_json()

        # validate request body
        errors = recover_pw_schema.validate(body)
        if errors:
            return {
                "msg": "Request body is incorrect",
                "errors": errors,
            }, HTTPStatus.BAD_REQUEST

        # obtain username/email from request
        username = body.get("username")
        email = body.get("email")

        log.debug("Scheduling handling of password recovery request")
        # we schedule _handle_password_recovery in '3' seconds to make it very
        # likely we'll respond to the user's request (HTTP) before we start
        # executing its code. We do this to avoid potential timing attacks
        gevent.spawn_later(
            3,
            _handle_password_recovery,
            current_app._get_current_object(),
            username,
            email,
            self.config,
            self.mail,
        )

        return ret


class ResetTwoFactorSecret(ServicesResources):
    """User can use recover token to reset their two-factor authentication."""

    def post(self):
        """Set a new two-factor authentication secret using a recover token
        ---
        description: >-
          User can use a recover token to reset their two-factor authentication
          secret

        requestBody:
          content:
            application/json:
              schema:
                properties:
                  reset_token:
                    type: string
                    description: Recover token (received by email)

        responses:
          200:
            description: Ok
          400:
            description: Recovery token is missing or invalid

        tags: ["Account recovery"]
        """
        # retrieve user based on email or username
        body = request.get_json()

        # validate request body
        errors = reset_2fa_schema.validate(body)
        if errors:
            return {
                "msg": "Request body is incorrect",
                "errors": errors,
            }, HTTPStatus.BAD_REQUEST

        # obtain user
        reset_token = body.get("reset_token")
        try:
            user_id = decode_token(reset_token)["sub"].get("id")
        except DecodeError:
            return {"msg": "Invalid recovery token!"}, HTTPStatus.BAD_REQUEST

        user = db.User.get(user_id)
        server_name = self.config.get("server_name", MAIN_VERSION_NAME)

        log.info("Resetting two-factor authentication for %s", user.username)
        return create_qr_uri(user, server_name), HTTPStatus.OK


class RecoverTwoFactorSecret(ServicesResources):
    """Send a mail containing a recover token for the 2fa secret"""

    def post(self):
        """Request a recover token to reset two-factor authentication secret
        ---
        description: >-
          Request a recover token if two-factor authentication secret is lost.
          A password and a username must be supplied.

        requestBody:
          content:
            application/json:
              schema:
                properties:
                  username:
                    type: string
                    description: Username from which the 2fa needs to be reset
                  password:
                    type: string
                    description: Password of user whose 2fa needs to be reset

        responses:
          200:
            description: Ok
          400:
            description: No username or email and password provided

        tags: ["Account recovery"]
        """
        # obtain parameters from request
        body = request.get_json()

        # validate request body
        errors = recover_2fa_schema.validate(body)
        if errors:
            return {
                "msg": "Request body is incorrect",
                "errors": errors,
            }, HTTPStatus.BAD_REQUEST

        username = body.get("username")
        password = body.get("password")

        # check credentials
        user, login_status = user_login(self.config, username, password, self.mail)
        if login_status != HTTPStatus.OK:
            log.error("Failed attempt to reset 2FA for submitted user '%s'", username)
            # Note: user_login() returns a dict with an error message if login
            #       failed as first returned element ('user')
            return user, login_status

        log.info("2FA reset requested for '%s'", user.username)

        # generate a token that can reset their password
        smtp_settings = self.config.get("smtp")
        if not smtp_settings:
            log.warning("No SMTP settings found in config - cannot send email!")
            return {
                "msg": "No mailserver settings found in server configuration - cannot "
                "send you an email to reset your 2FA! Contact your administrator."
            }, HTTPStatus.BAD_REQUEST
        minutes_token_valid = smtp_settings.get(
            "email_token_validity_minutes", DEFAULT_EMAILED_TOKEN_VALIDITY_MINUTES
        )
        expires = dt.timedelta(minutes=minutes_token_valid)
        reset_token = create_access_token({"id": str(user.id)}, expires_delta=expires)

        email_from = smtp_settings.get("email_from", DEFAULT_EMAIL_FROM_ADDRESS)
        support_email = self.config.get("support_email", DEFAULT_SUPPORT_EMAIL_ADDRESS)

        self.mail.send_email(
            f"Two-factor authentication reset {APPNAME}",
            sender=email_from,
            recipients=[user.email],
            text_body=render_template(
                "mail/reset_token.txt",
                token=reset_token,
                firstname=user.firstname,
                reset_type="two-factor authentication code",
                what_to_do=("please reset your password! It has been compromised"),
            ),
            html_body=render_template(
                "mail/reset_token.html",
                token=reset_token,
                firstname=user.firstname,
                reset_type="two-factor authentication code",
                support_email=support_email,
                what_to_do=("please reset your password! It has been compromised"),
            ),
        )
        log.info("2FA reset request email sent for '%s'", user.username)

        return {
            "msg": "You should have received an email that will allow you to reset your"
            " 2FA."
        }, login_status


class ChangePassword(ServicesResources):
    """
    Let user change their password with old password as verification
    """

    @with_user
    def patch(self):
        """Set a new password using the current password
        ---
        description: >-
          Users can change their password by submitting their current password
          and a new password

        requestBody:
          content:
            application/json:
              schema:
                properties:
                  current_password:
                    type: string
                    description: current password
                  new_password:
                    type: string
                    description: new password

        responses:
          200:
            description: Ok
          400:
            description: Current or new password is missing from JSON body, or
              they are the same, or the new password doesn't meet password
              criteria
          401:
            description: Current password is incorrect

        tags: ["Account recovery"]
        """
        body = request.get_json()
        # validate request body
        errors = change_pw_schema.validate(body)
        if errors:
            return {
                "msg": "Request body is incorrect",
                "errors": errors,
            }, HTTPStatus.BAD_REQUEST

        old_password = body.get("current_password")
        new_password = body.get("new_password")

        user: User = g.user
        log.info("Changing password for user %s", user.id)

        # check if the old password is correct
        user_or_error, code = user_login(
            self.config, user.username, old_password, self.mail
        )
        if code != HTTPStatus.OK:
            return user_or_error, code

        if old_password == new_password:
            return {
                "msg": "New password is the same as current password!"
            }, HTTPStatus.BAD_REQUEST

        # set password
        msg = user.set_password(new_password)
        if msg:
            return {"msg": msg}, HTTPStatus.BAD_REQUEST

        log.info("Successful password change for '%s'", user.username)
        return {"msg": "The password has been changed successfully!"}, HTTPStatus.OK


class ResetAPIKey(ServicesResources):
    """User can reset API key."""

    def __init__(self, socketio, storage_adapter, mail, api, permissions, config):
        super().__init__(socketio, storage_adapter, mail, api, permissions, config)

        # obtain permissions to check if user is allowed to modify nodes
        self.r = getattr(self.permissions, "node")

    @with_user
    def post(self):
        """Reset a node's API key
        ---
        description: >-
            If a node's API key is lost, this route can be used to obtain a new
            API key.\n

            ### Permission Table\n
            |Rule name|Scope|Operation|Assigned to node|Assigned to container|
            Description|\n
            |--|--|--|--|--|--|\n
            |Node|Global|Edit|❌|❌|Reset API key of node specified by id|\n
            |Node|Organization|Edit|❌|❌|Reset API key of node specified by
            id which is part of your organization |\n

            Accessible to users.

        requestBody:
          content:
            application/json:
              schema:
                properties:
                  id:
                    type: integer
                    description: ID of node whose API key is to be reset

        responses:
            200:
                description: Ok
            400:
                description: ID missing from json body
            401:
                description: Unauthorized
            404:
                description: Node not found

        security:
            - bearerAuth: []

        tags: ["Account recovery"]
        """
        body = request.get_json()

        # validate request body
        errors = reset_api_key_schema.validate(body)
        if errors:
            return {
                "msg": "Request body is incorrect",
                "errors": errors,
            }, HTTPStatus.BAD_REQUEST

        id_ = body["id"]
        node = db.Node.get(id_)
        if not node:
            return {"msg": f"Node id={id_} is not found!"}, HTTPStatus.NOT_FOUND

        # check if user is allowed to edit the node
        if not self.r.allowed_for_org(Operation.EDIT, node.organization_id):
            return {
                "msg": "You lack the permission to do that!"
            }, HTTPStatus.UNAUTHORIZED

        # all good, change API key
        log.info("Successful API key reset for node %s", id_)
        api_key = generate_apikey()
        node.api_key = api_key
        node.save()

        return {"api_key": api_key}, HTTPStatus.OK



from pathlib import Path

from vantage6.common.globals import APPNAME

#
#   INSTALLATION SETTINGS
#
PACKAGE_FOLDER = Path(__file__).parent.parent.parent

DATA_FOLDER = PACKAGE_FOLDER / APPNAME / "server" / "_data"

SERVER_MODULE_NAME = APPNAME + "-server"

#
#   RUNTIME SETTINGS
#

# Expiretime of JWT tokens
ACCESS_TOKEN_EXPIRES_HOURS = 6

# minimum validity of JWT Tokens in seconds
MIN_TOKEN_VALIDITY_SECONDS = 1800

# Expiration time of refresh tokens
REFRESH_TOKENS_EXPIRE_HOURS = 48

# Minimum time in seconds that a refresh token must be valid *longer than* the
# access token. This is to prevent the access token from expiring before the
# refresh token.
MIN_REFRESH_TOKEN_EXPIRY_DELTA = 1

# Where the resources modules have to be loaded from
RESOURCES_PATH = "vantage6.server.resource"

# Which resources should be initialized. These names correspond to the
# file-names in the resource directory
RESOURCES = [
    "node",
    "collaboration",
    "organization",
    "task",
    "run",
    "blobstream",
    "token",
    "user",
    "version",
    "recover",
    "role",
    "rule",
    "health",
    "vpn",
    "port",
    "event",
    "ui.column",
    "algorithm_store",
    "study",
]

# Super user information. This user is only created if it is not in the
# database yet at startup time.
SUPER_USER_INFO = {"username": "root", "password": "root"}

# default time that token is valid in minutes
DEFAULT_EMAILED_TOKEN_VALIDITY_MINUTES = 60

# default password policies
DEFAULT_MAX_FAILED_ATTEMPTS = 5
DEFAULT_INACTIVATION_MINUTES = 15
DEFAULT_BETWEEN_USER_EMAILS_MINUTES = 60



import logging
import base64

from marshmallow import fields
from flask import url_for

from vantage6.server import db
from vantage6.common import logger_name
from vantage6.common.globals import STRING_ENCODING
from vantage6.server.model import User
from vantage6.backend.common.resource.output_schema import (
    BaseHATEOASModelSchema,
    create_one_to_many_link,
)

log = logging.getLogger(logger_name(__name__))


class HATEOASModelSchema(BaseHATEOASModelSchema):
    """
    This class is used to convert foreign-key fields to HATEOAS specification.
    """

    def __init__(self, *args, **kwargs) -> None:
        # set lambda functions to create links for one to one relationship
        setattr(self, "node", lambda obj: self.create_hateoas("node", obj))
        setattr(
            self, "organization", lambda obj: self.create_hateoas("organization", obj)
        )
        setattr(
            self, "collaboration", lambda obj: self.create_hateoas("collaboration", obj)
        )
        setattr(self, "user", lambda obj: self.create_hateoas("user", obj))
        setattr(self, "run", lambda obj: self.create_hateoas("run", obj))
        setattr(self, "task", lambda obj: self.create_hateoas("task", obj))
        setattr(self, "port", lambda obj: self.create_hateoas("port", obj))
        setattr(
            self,
            "parent_",
            lambda obj: self.create_hateoas("parent", obj, endpoint="task"),
        )
        setattr(
            self,
            "init_org_",
            lambda obj: self.create_hateoas("init_org", obj, endpoint="organization"),
        )
        setattr(
            self,
            "init_user_",
            lambda obj: self.create_hateoas("init_user", obj, endpoint="user"),
        )
        setattr(self, "study", lambda obj: self.create_hateoas("study", obj))
        setattr(
            self,
            "algorithm_store",
            lambda obj: self.create_hateoas("algorithm_store", obj),
        )

        # call super class. Do this after setting the attributes above, because
        # the super class initializer will call the attributes.
        super().__init__(*args, **kwargs)


# /task/{id}
class TaskSchema(HATEOASModelSchema):
    class Meta:
        model = db.Task

    status = fields.String()
    finished_at = fields.DateTime()
    collaboration = fields.Method("collaboration")
    runs = fields.Function(
        lambda obj: create_one_to_many_link(obj, link_to="run", link_from="task_id")
    )
    results = fields.Function(
        lambda obj: create_one_to_many_link(obj, link_to="result", link_from="task_id")
    )
    parent = fields.Method("parent_")
    children = fields.Function(
        lambda obj: create_one_to_many_link(obj, link_to="task", link_from="parent_id")
    )
    init_org = fields.Method("init_org_")
    init_user = fields.Method("init_user_")
    databases = fields.Method("databases_")
    study = fields.Method("study")
    algorithm_store = fields.Method("algorithm_store")

    @staticmethod
    def databases_(obj):
        return [
            {"label": db.database, "parameters": db.parameters} for db in obj.databases
        ]


class ResultSchema(HATEOASModelSchema):
    class Meta:
        model = db.Run
        exclude = (
            "assigned_at",
            "started_at",
            "finished_at",
            "status",
            "ports",
            "organization",
            "log",
            "input",
        )

    run = fields.Method("make_run_link")
    task = fields.Method("task")

    @staticmethod
    def make_run_link(obj):
        return {
            "id": obj.id,
            "link": url_for("run_with_id", id=obj.id),
            "methods": ["GET", "PATCH"],
        }


# /task/{id}?include=runs
class TaskWithRunSchema(TaskSchema):
    """Returns the TaskSchema plus the correspoding runs."""

    runs = fields.Nested("RunSchema", many=True)


# /task/{id}?include=results
class TaskWithResultSchema(TaskSchema):
    """Returns the TaskSchema plus the correspoding results."""

    results = fields.Nested("ResultSchema", many=True)


# /task/{id}?include=runs,results
class TaskWithRunAndResultSchema(TaskSchema):
    """Returns the TaskSchema plus the correspoding runs and results."""

    runs = fields.Nested("RunSchema", many=True)
    results = fields.Nested("ResultSchema", many=True)


class RunSchema(HATEOASModelSchema):
    class Meta:
        model = db.Run
        exclude = ("result",)

    organization = fields.Method("organization")
    task = fields.Method("task")
    results = fields.Method("result_link")
    node = fields.Function(
        serialize=lambda obj: RunNodeSchema().dump(obj.node, many=False)
    )
    ports = fields.Function(
        serialize=lambda obj: RunPortSchema().dump(obj.ports, many=True)
    )

    @staticmethod
    def result_link(obj):
        return {
            "id": obj.id,
            "link": url_for("result_with_id", id=obj.id),
            "methods": ["GET", "PATCH"],
        }


class RunTaskIncludedSchema(RunSchema):
    task = fields.Nested("TaskSchema", many=False, exclude=["runs"])


class RunNodeSchema(HATEOASModelSchema):
    class Meta:
        model = db.Node
        exclude = ("type", "api_key", "collaboration", "organization", "last_seen")


class PortSchema(HATEOASModelSchema):
    class Meta:
        model = db.AlgorithmPort


class RunPortSchema(HATEOASModelSchema):
    class Meta:
        model = db.AlgorithmPort
        exclude = ("run",)


class OrganizationSchema(HATEOASModelSchema):
    class Meta:
        model = db.Organization
        exclude = ("_public_key",)

    # add links to linked resources
    collaborations = fields.Function(
        lambda obj: create_one_to_many_link(
            obj, link_to="collaboration", link_from="organization_id"
        )
    )
    nodes = fields.Function(
        lambda obj: create_one_to_many_link(
            obj, link_to="node", link_from="organization_id"
        )
    )
    users = fields.Function(
        lambda obj: create_one_to_many_link(
            obj, link_to="user", link_from="organization_id"
        )
    )
    tasks = fields.Function(
        lambda obj: create_one_to_many_link(
            obj, link_to="task", link_from="init_org_id"
        )
    )
    runs = fields.Function(
        lambda obj: create_one_to_many_link(
            obj, link_to="run", link_from="organization_id"
        )
    )
    studies = fields.Function(
        lambda obj: create_one_to_many_link(
            obj, link_to="study", link_from="organization_id"
        )
    )

    public_key = fields.Function(
        lambda obj: (
            base64.b64encode(obj._public_key).decode(STRING_ENCODING)
            if obj._public_key
            else ""
        )
    )


class CollaborationSchema(HATEOASModelSchema):
    class Meta:
        model = db.Collaboration

    # add links to linked resources
    organizations = fields.Function(
        lambda obj: create_one_to_many_link(
            obj, link_to="organization", link_from="collaboration_id"
        )
    )
    nodes = fields.Function(
        lambda obj: create_one_to_many_link(
            obj, link_to="node", link_from="collaboration_id"
        )
    )
    tasks = fields.Function(
        lambda obj: create_one_to_many_link(
            obj, link_to="task", link_from="collaboration_id"
        )
    )
    algorithm_stores = fields.Function(
        lambda obj: create_one_to_many_link(
            obj, link_to="algorithm_store", link_from="collaboration_id"
        )
    )
    studies = fields.Function(
        lambda obj: create_one_to_many_link(
            obj, link_to="study", link_from="collaboration_id"
        )
    )


class CollaborationWithOrgsSchema(CollaborationSchema):
    """
    Returns the CollaborationSchema plus the organizations participating in it.
    """

    organizations = fields.Nested("OrganizationSchema", many=True)


class StudySchema(HATEOASModelSchema):
    class Meta:
        model = db.Study

    collaboration = fields.Method("collaboration")
    organizations = fields.Function(
        lambda obj: create_one_to_many_link(
            obj, link_to="organization", link_from="study_id"
        )
    )
    tasks = fields.Function(
        lambda obj: create_one_to_many_link(obj, link_to="task", link_from="study_id")
    )


class StudyWithOrgsSchema(StudySchema):
    """
    Returns the StudySchema plus the organizations participating in it.
    """

    organizations = fields.Nested("OrganizationSchema", many=True)


class NodeSchema(HATEOASModelSchema):
    organization = fields.Method("organization")
    collaboration = fields.Method("collaboration")
    config = fields.Nested("NodeConfigSchema", many=True, exclude=["id"])

    class Meta:
        model = db.Node
        exclude = ("api_key",)


class NodeConfigSchema(HATEOASModelSchema):
    class Meta:
        model = db.NodeConfig


class NodeSchemaSimple(HATEOASModelSchema):
    organization = fields.Method("organization")

    class Meta:
        model = db.Node
        exclude = (
            "collaboration",
            "api_key",
            "type",
        )


class AlgorithmStoreSchema(HATEOASModelSchema):
    class Meta:
        model = db.AlgorithmStore

    collaborations = fields.Function(
        lambda obj: (
            create_one_to_many_link(
                obj, link_to="collaboration", link_from="algorithm_store_id"
            )
            if obj.collaboration_id
            else None
        )
    )


class UserSchema(HATEOASModelSchema):
    class Meta:
        model = db.User
        exclude = ("password", "failed_login_attempts", "last_login_attempt")

    roles = fields.Function(
        lambda obj: create_one_to_many_link(obj, link_to="role", link_from="user_id")
    )
    rules = fields.Function(
        lambda obj: create_one_to_many_link(obj, link_to="rule", link_from="user_id")
    )

    organization = fields.Method("organization")


class UserWithPermissionDetailsSchema(UserSchema):
    """
    A schema for API responses that contains regular user details plus
    additional permission details for the user to be used by the UI.
    """

    permissions = fields.Method("permissions_")

    @staticmethod
    def permissions_(obj: User) -> dict:
        """
        Returns a dictionary containing permission details for the user to be
        used by the UI.

        Parameters
        ----------
        obj : User
            The user to get permission details for.

        Returns
        -------
        dict
            A dictionary containing permission details for the user.
        """
        role_ids = [role.id for role in obj.roles]
        rule_ids = list(
            set(
                [rule.id for rule in obj.rules]
                + [rule.id for role in obj.roles for rule in role.rules]
            )
        )
        # Return which organizations the user is in collaboration with so that
        # UI knows which organizations user has access to if they have
        # collaboration scope permissions
        orgs_in_collabs = list(
            set(
                [
                    org.id
                    for collab in obj.organization.collaborations
                    for org in collab.organizations
                ]
            )
        )
        return {
            "roles": role_ids,
            "rules": rule_ids,
            "orgs_in_collabs": orgs_in_collabs,
        }


class RoleSchema(HATEOASModelSchema):
    rules = fields.Function(
        lambda obj: create_one_to_many_link(obj, link_to="rule", link_from="role_id")
    )
    users = fields.Function(
        lambda obj: create_one_to_many_link(obj, link_to="user", link_from="role_id")
    )
    organization = fields.Method("organization")

    class Meta:
        model = db.Role


class RuleSchema(HATEOASModelSchema):
    scope = fields.Function(serialize=lambda obj: obj.scope.name)
    operation = fields.Function(serialize=lambda obj: obj.operation.name)
    users = fields.Function(
        lambda obj: create_one_to_many_link(obj, link_to="user", link_from="rule_id")
    )

    class Meta:
        model = db.Rule
        exclude = ("roles",)



import datetime as dt


def parse_datetime(
    date: str | dt.datetime = None, default: dt.datetime = None
) -> dt.datetime:
    """
    Utility function to parse a datetime string.

    Parameters
    ----------
    date : str | datetime.datetime, optional
        Datetime string
    default : datetime.datetime, optional
        Default datetime to return if `dt` is None

    Returns
    -------
    datetime.datetime
        Datetime object
    """
    if date:
        if isinstance(date, str):
            converter = "%Y-%m-%dT%H:%M:%S.%f"
            if date.endswith("+00:00"):
                converter += "%z"  # parse timezone
            return dt.datetime.strptime(date, converter)
        else:
            # convert datetime to UTC
            return date.astimezone(dt.timezone.utc)
    return default



import bcrypt


class HashedPassword(str):
    """
    Represents a bcrypt hashed password in string form. Used for allowing users
    to provide a hashed initial password during server setup.
    """

    def __new__(cls, value: str):
        # attempt to use provided value as a bcrypt hash as a check
        if not cls._looks_valid_bcrypt(value):
            raise ValueError("Provided value does not appear to be a valid bcrypt hash")
        return super().__new__(cls, value)

    @staticmethod
    def _looks_valid_bcrypt(value: str):
        """
        Check if the given value *looks* like a potentially valid bcrypt hash
        by attempting to check it with bcrypt.checkpw(). Also checks for hashed
        empty string.

        `bcrypt.checkpw()` will raise ValueError if the value is not a valid,
        can also raise PanicException if not enough length.

        Parameters
        ----------
        value : str
            Purported bcrypt hash

        Returns
        -------
        bool
            True if the value looks like a valid bcrypt hash. False otherwise,
            or if the empty string hashes to the given value.
        """
        if len(value) < 60:
            return False

        # we enforce $2b$ version
        if not value.startswith("$2b$"):
            return False

        try:
            empty_check = bcrypt.checkpw(b"", value.encode("utf8"))
            # the empty string hashes to the given 'hashed password' we
            # consider this invalid out of caution
            if empty_check:
                return False
        except ValueError:
            return False

        return True



import logging
from typing import IO, Union

from azure.identity import ClientSecretCredential
from azure.storage.blob import BlobServiceClient
from sqlalchemy import event

from vantage6.common import logger_name
from vantage6.server.model.run import Run

module_name = logger_name(__name__)
log = logging.getLogger(module_name)


class AzureStorageService:
    """
    A service for managing Azure Blob Storage.
    """

    def __init__(self, config: dict):
        """
        Initialize the AzureStorageService.

        Parameters
        ----------
        config : dict
            Configuration dictionary containing Azure Blob Storage settings.
        """
        self.blob_service_client = None

        tenant_id = config.get("tenant_id")
        client_id = config.get("client_id")
        client_secret = config.get("client_secret")
        storage_account_name = config.get("storage_account_name")
        container_name = config.get("container_name")
        connection_string = config.get("connection_string")

        if tenant_id and client_id and client_secret and storage_account_name:
            credential = ClientSecretCredential(
                tenant_id=tenant_id, client_id=client_id, client_secret=client_secret
            )
            self.blob_service_client = BlobServiceClient(
                account_url=f"https://{storage_account_name}.blob.core.windows.net/",
                credential=credential,
            )

        if connection_string:
            self.blob_service_client = BlobServiceClient.from_connection_string(
                connection_string
            )

        if not self.blob_service_client:
            log.warning(
                "Azure Blob Storage configuration is incomplete. Large result store not set up."
            )
            return

        if container_name:
            self.container_name = container_name
        else:
            raise ValueError("Container name must be provided.")

        self.container_client = self.blob_service_client.get_container_client(
            container_name
        )
        event.listen(Run, "after_delete", self.delete_blob_after_run_delete)

    def get_blob(self, blob_name: str) -> bytes:
        """
        Retrieve a blob from Azure Blob Storage by its name.

        Parameters
        ----------
        blob_name : str
            The name of the blob to retrieve.

        Returns
        -------
        bytes
            The content of the blob.
        """
        log.debug(f"Retrieving blob: {blob_name} from container: {self.container_name}")
        blob_client = self.blob_service_client.get_blob_client(
            container=self.container_name, blob=blob_name
        )
        stream = blob_client.download_blob()
        return stream.readall()

    def store_blob(self, blob_name: str, data: Union[IO, bytes]) -> None:
        """
        Store data as a blob in Azure Blob Storage.

        Parameters
        ----------
        blob_name : str
            The name of the blob to create or overwrite.
        data : Union[IO, bytes]
            The data to store in the blob. Can be a bytes object or a file-like
            object.
        """
        log.debug(f"Storing blob: {blob_name} in container: {self.container_name}")
        blob_client = self.blob_service_client.get_blob_client(
            container=self.container_name, blob=blob_name
        )
        try:
            blob_client.upload_blob(data, overwrite=True)
        except Exception as e:
            log.error(f"Failed to upload blob '{blob_name}': {e}")
            raise RuntimeError(f"Failed to upload blob '{blob_name}': {e}")

    def delete_blob(self, blob_name: str) -> None:
        """
        Delete a blob from Azure Blob Storage by its name.

        Parameters
        ----------
        blob_name : str
            The name of the blob to delete.
        """
        log.debug(f"Deleting blob: {blob_name} from container: {self.container_name}")
        blob_client = self.blob_service_client.get_blob_client(
            container=self.container_name, blob=blob_name
        )
        blob_client.delete_blob()

    def stream_blob(self, blob_name: str):
        """
        Stream a blob from Azure Blob Storage.
        Returns a StorageStreamDownloader object.

        Parameters
        ----------
        blob_name : str
            The name of the blob to stream.

        Returns
        -------
        StorageStreamDownloader
            A stream object to read the blob's content in chunks.
        """
        log.debug(f"Streaming blob: {blob_name} from container: {self.container_name}")
        blob_client = self.blob_service_client.get_blob_client(
            container=self.container_name, blob=blob_name
        )
        return blob_client.download_blob()

    def delete_blob_after_run_delete(self, mapper, connection, target):
        """
        SQLAlchemy event listener to delete the associated blob when a Run
        instance is deleted.
        """
        if target.blob_storage_used:
            try:
                if target.result:
                    self.delete_blob(target.result)
                if target.input:
                    self.delete_blob(target.input)
            except Exception as e:
                error_msg = f"Failed to delete blob for run {target.id}: {e}"
                log.error(error_msg)
                raise RuntimeError(error_msg)



import unittest

from vantage6.server.hashedpassword import HashedPassword


class TestHashedPassword(unittest.TestCase):
    def test_valid_hash(self):
        # hash 'vantage6test'
        HashedPassword("$2b$12$qywY/MVIAVm6YUPfAWZR5OBUCWFEF58FwdKas7Mhcb.HkICkXAzUe")

    def test_invalid_hash_length(self):
        # missing last character
        with self.assertRaises(ValueError):
            HashedPassword(
                "$2b$12$qywY/MVIAVm6YUPfAWZR5OBUCWFEF58FwdKas7Mhcb.HkICkXAzU"
            )

    def test_invalid_hash(self):
        # didn't understand the assignment
        with self.assertRaises(ValueError):
            HashedPassword("superpassword")

    def test_empty_hash(self):
        # empty bcrypt hash
        with self.assertRaises(ValueError):
            HashedPassword("")

    def test_empty_string_hash(self):
        # hash of empty string
        with self.assertRaises(ValueError):
            HashedPassword(
                "$2b$12$iMwnJmCzBO0Jki.Lu4Qfmuv80L7FGbLLAPRYIxzWUhjnjbG/xxCjm"
            )

    def test_not_bcrypt_hash(self):
        # not a bcrypt hash of 'vantage6test' (openssl passwd -6)
        with self.assertRaises(ValueError):
            HashedPassword(
                "$6$8HxGnDKoIDorOIyk$Fj.lInEmIW.XTMYBmWKB7w.s7ZwKDpy3wHCJCTPv7O0tA00jGlYXlLVB/61FgnToyFp5IdJOq1ze3QJU5g1zz."
            )

    def test_older_version(self):
        # 2a version hashing 'vantage6test'
        with self.assertRaises(ValueError):
            HashedPassword(
                "$2a$12$4SBnOHg0PqtzLQu/vN0.4un3N21G3tJazxBuu9hJSi0WbF9LRONdm"
            )



from __future__ import annotations
from sqlalchemy import Column, String, Integer, ForeignKey
from sqlalchemy.orm import relationship

from vantage6.server.model.base import Base


class Study(Base):
    """
    Table that describes which studies are available.

    Studies are subsets of one or more organizations within a collaboration that
    collaborate on a particular research question. Studies use the same nodes and data
    as the collaboration they are part of, and also inherit their collaboration's
    settings.

    Attributes
    ----------
    name : str
        Name of the study
    collaboration : :class:`~vantage6.server.model.collaboration.Collaboration`
        Collaboration that this study is part of
    organizations :
            list[:class:`~vantage6.server.model.organization.Organization`]
        List of organizations that are part of this study
    tasks : list[:class:`~vantage6.server.model.task.Task`]
        List of tasks that are part of this study
    """

    # fields
    name = Column(String, unique=True)
    collaboration_id = Column(Integer, ForeignKey("collaboration.id"))

    # relationships
    collaboration = relationship("Collaboration", back_populates="studies")
    organizations = relationship(
        "Organization", secondary="StudyMember", back_populates="studies"
    )
    tasks = relationship("Task", back_populates="study")

    def __repr__(self):
        """
        Returns a string representation of the study.

        Returns
        -------
        str
            String representation of the study
        """
        number_of_organizations = len(self.organizations)
        number_of_tasks = len(self.tasks)
        return (
            "<Study "
            f"{self.id}: '{self.name}', "
            f"collaboration_id={self.collaboration_id}, "
            f"{number_of_organizations} organization(s), "
            f"{number_of_tasks} task(s)"
            ">"
        )



import codecs
import os

from os import path
from setuptools import setup, find_namespace_packages
from pathlib import Path

# get current directory
here = Path(path.abspath(path.dirname(__file__)))
parent_dir = here.parent.absolute()

# get the long description from the README file
with codecs.open(path.join(parent_dir, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

# Read the API version from disk. This file should be located in the package
# folder, since it's also used to set the pkg.__version__ variable.
version_path = os.path.join(here, "vantage6", "server", "_version.py")
version_ns = {"__file__": version_path}
with codecs.open(version_path) as f:
    exec(f.read(), {}, version_ns)

# setup the package
setup(
    name="vantage6_server",
    version=version_ns["__version__"],
    description="Vantage6 server",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/vantage6/vantage6",
    packages=find_namespace_packages(),
    python_requires=">=3.10",
    install_requires=[
        "azure-identity>=1.22.0",
        "azure-storage-blob>=12.25.1",
        "bcrypt==4.0.1",
        "flask==3.1.1",
        "flask-cors==6.0.0",
        "Flask-JWT-Extended==4.7.1",
        "Flask-Principal==0.4.0",
        "Flask-RESTful==0.3.10",
        "flask-marshmallow==0.15.0",
        "flask-socketio==5.5.1",
        "gevent==25.4.2",
        "ipython==8.10.0",
        "kombu==5.2.4",
        "marshmallow==3.26.2",
        "pyjwt==2.12.1",
        "pyotp==2.8.0",
        "prometheus-client==0.21.1",
        "requests==2.32.4",
        "requests-oauthlib==1.3.1",
        "schema==0.7.5",
        "SQLAlchemy==1.4.46",
        "werkzeug==3.1.5",
        f"vantage6 == {version_ns['__version__']}",
        f"vantage6-common == {version_ns['__version__']}",
    ],
    extras_require={
        "dev": [
            "coverage==6.4.4",
            "black",
            "pre-commit",
        ]
    },
    package_data={
        "vantage6.server": [
            "__build__",
            "_data/**/*.yaml",
            "server_data/*.yaml",
        ],
    },
    entry_points={
        "console_scripts": ["vserver-local=vantage6.server.cli.server:cli_server"]
    },
)
